"""A Python based FEM solver.
Copyright (C) 2025  name of Robert Fennis

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, see
<https://www.gnu.org/licenses/>.

"""
import os

NTHREADS = "1"

os.environ["OMP_NUM_THREADS"] = NTHREADS
os.environ["MKL_NUM_THREADS"] = NTHREADS
os.environ["OPENBLAS_NUM_THREADS"] = NTHREADS
os.environ["VECLIB_MAXIMUM_THREADS"] = NTHREADS
os.environ["NUMEXPR_NUM_THREADS"] = NTHREADS


from loguru import logger
from _emerge.logsettings import logger_format
import sys

logger.remove()
logger.add(sys.stderr, format=logger_format)

logger.debug('Importing modules')

from _emerge.simmodel import Simulation3D
from _emerge.material import Material
from _emerge import bc
from _emerge.solver import SolverBicgstab, SolverGMRES, SolveRoutine, ReverseCuthillMckee, Sorter, SolverPardiso, SolverUMFPACK, SolverSuperLU, EMSolver
from _emerge.cs import CoordinateSystem, CS, GCS, Plane, Axis, XAX, YAX, ZAX, XYPLANE, XZPLANE, YZPLANE, YXPLANE, ZXPLANE, ZYPLANE
from _emerge.coord import Line
from _emerge import geo
from _emerge.selection import Selection, FaceSelection, DomainSelection, EdgeSelection
from _emerge.mth.common_functions import norm, coax_rout, coax_rin
from _emerge.physics.microwave.sc import stratton_chu
from _emerge.periodic import RectCell, HexCell
from _emerge.mesher import Algorithm2D, Algorithm3D
from . import lib
from _emerge.howto import _HowtoClass
howto = _HowtoClass()
logger.debug('Importing complete!')


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.


from __future__ import annotations
from scipy.sparse import lil_matrix, csc_matrix
from scipy.sparse.csgraph import reverse_cuthill_mckee
from scipy.sparse.linalg import bicgstab, gmres, gcrotmk, eigs, splu
from scipy.linalg import eig
from scipy import sparse
from dataclasses import dataclass
import numpy as np
from loguru import logger
import platform
import time
from typing import Literal
from enum import Enum

_PARDISO_AVAILABLE = False
_UMFPACK_AVAILABLE = False
_PARDISO_ERROR_CODES = """
   0 | No error.
  -1 | Input inconsistent.
  -2 | Not enough memory.
  -3 | Reordering problem.
  -4 | Zero pivot, numerical fac. or iterative refinement problem.
  -5 | Unclassified (internal) error.
  -6 | Preordering failed (matrix types 11(real and nonsymmetric), 13(complex and nonsymmetric) only).
  -7 | Diagonal Matrix problem.
  -8 | 32-bit integer overflow problem.
 -10 | No license file pardiso.lic found.
 -11 | License is expired.
 -12 | Wrong username or hostname.
-100 | Reached maximum number of Krylov-subspace iteration in iterative solver.
-101 | No sufficient convergence in Krylov-subspace iteration within 25 iterations.
-102 | Error in Krylov-subspace iteration.
-103 | Bread-Down in Krylov-subspace iteration
"""
""" Check if the PC runs on a non-ARM architechture
If so, attempt to import PyPardiso (if its installed)
"""

if 'arm' not in platform.processor():
    try:
        from pypardiso import spsolve as pardiso_solve
        from pypardiso.pardiso_wrapper import PyPardisoError
        _PARDISO_AVAILABLE = True
    except ModuleNotFoundError as e:
        logger.info('Pardiso not found, defaulting to SuperLU')
try:
    import scikits.umfpack as um
    _UMFPACK_AVAILABLE = True
except ModuleNotFoundError as e:
    logger.debug('UMFPACK not found, defaulting to SuperLU')

def filter_real_modes(eigvals, eigvecs, k0, ermax, urmax, sign):
    """
    Given arrays of eigenvalues `eigvals` and eigenvectors `eigvecs` (cols of shape (N,)),
    and a free‐space wavenumber k0, return only those eigenpairs whose eigenvalue can
    correspond to a real propagation constant β (i.e. 0 ≤ β² ≤ k0²·ermax·urmax).

    Assumes that `ermax` and `urmax` are defined in the surrounding scope.

    Parameters
    ----------
    eigvals : 1D array_like of float
        The generalized eigenvalues (β² candidates).
    eigvecs : 2D array_like, shape (N, M)
        The corresponding eigenvectors, one column per eigenvalue.
    k0 : float
        Free‐space wavenumber.

    Returns
    -------
    filtered_vals : 1D ndarray
        Subset of `eigvals` satisfying 0 ≤ eigval ≤ k0²·ermax·urmax (within numerical tol).
    filtered_vecs : 2D ndarray
        Columns of `eigvecs` corresponding to `filtered_vals`.
    """
    minimum = 1
    extremum = (k0**2) * ermax * urmax * 2
    
    mask = (sign*eigvals <= extremum) & (sign*eigvals >= minimum)
    filtered_vals = eigvals[mask]
    filtered_vecs = eigvecs[:, mask]
    k0vals = np.sqrt(sign*filtered_vals)
    order = np.argsort(np.abs(k0vals))# ascending distance
    filtered_vals = filtered_vals[order]             # reorder eigenvalues
    filtered_vecs = filtered_vecs[:, order] 
    return filtered_vals, filtered_vecs

def filter_unique_eigenpairs(eigen_values: list[complex], eigen_vectors: list[np.ndarray], tol=-3) -> tuple[list[complex], list[np.ndarray]]:
    """
    Filters eigenvectors by orthogonality using dot-product tolerance.
    
    Parameters:
        eigen_values (np.ndarray): Array of eigenvalues, shape (n,)
        eigen_vectors (np.ndarray): Array of eigenvectors, shape (n, n)
        tol (float): Dot product tolerance for considering vectors orthogonal (default: 1e-5)

    Returns:
        unique_values (np.ndarray): Filtered eigenvalues
        unique_vectors (np.ndarray): Corresponding orthogonal eigenvectors
    """
    selected = []
    indices = []
    for i in range(len(eigen_vectors)):
        
        vec = eigen_vectors[i]
        vec = vec / np.linalg.norm(vec)  # Normalize

        # Check orthogonality against selected vectors
        if all(10*np.log10(abs(np.dot(vec, sel))) < tol for sel in selected):
            selected.append(vec)
            indices.append(i)

    unique_values = [eigen_values[i] for i in indices]
    unique_vectors = [eigen_vectors[i] for i in indices]

    return unique_values, unique_vectors

def complex_to_real_block(A, b):
    """Return (Â,  b̂) real-augmented representation of A x = b."""
    A_r = sparse.csr_matrix(A.real)
    A_i = sparse.csr_matrix(A.imag)
    #  [ ReA  -ImA ]
    #  [ ImA   ReA ]
    upper = sparse.hstack([A_r, -A_i])
    lower = sparse.hstack([A_i,  A_r])
    A_hat = sparse.vstack([upper, lower]).tocsr()

    b_hat = np.hstack([b.real, b.imag])
    return A_hat, b_hat

def real_to_complex_block(x):
    """Return x = (x_r, x_i) as complex vector."""
    n = x.shape[0] // 2
    x_r = x[:n]
    x_i = x[n:]
    return x_r + 1j * x_i

    
class Sorter:
    """ A Generic class that executes a sort on the indices.
    It must implement a sort and unsort method.
    """
    def __init__(self):
        self.perm = None
        self.inv_perm = None

    def reset(self) -> str:
        """ Reset the permuation vectors."""
        self.perm = None
        self.inv_perm = None

    def __str__(self) -> str:
        return f'{self.__class__.__name__}'
    
    def sort(self, A: lil_matrix, b: np.ndarray, reuse_sorting: bool = False) -> tuple[lil_matrix, np.ndarray]:
        return A,b
    
    def unsort(self, x: np.ndarray) -> np.ndarray:
        return x

class Preconditioner:
    """A Generic class defining a preconditioner as attribute .M based on the
    matrix A and b. This must be generated in the .init(A,b) method.
    """
    def __init__(self):
        self.M: np.ndarray = None

    def __str__(self) -> str:
        return f'{self.__class__.__name__}'
    
    def init(self, A: lil_matrix, b: np.ndarray) -> None:
        pass

class Solver:
    """ A generic class representing a solver for the problem Ax=b
    
    A solver class has two class attributes.
     - real_only: defines if the solver can only deal with real numbers. In this case
    the solve routine will automatically provide A and b in real number format.
     - req_sorter: defines if this solver requires the use of a sorter algorithm. By setting
     it to False, the SolveRoutine will not use the default sorting algorithm.
    """
    real_only: bool = False
    req_sorter: bool = False
    def __init__(self):
        self.own_preconditioner: bool = False

    def __str__(self) -> str:
        return f'{self.__class__.__name__}'
    
    def solve(self, A: lil_matrix, b: np.ndarray, precon: Preconditioner, reuse_factorization: bool = False, id: int = -1) -> tuple[np.ndarray, int]:
        raise NotImplementedError("This classes Ax=B solver method is not implemented.")
    
    def reset(self) -> None:
        """Reset state variables like numeric and symbollic factorizations."""
        pass

class EigSolver:
    """ A generic class representing a solver for the eigenvalue problem Ax=λBx
    
    A solver class has two class attributes.
     - real_only: defines if the solver can only deal with real numbers. In this case
    the solve routine will automatically provide A and b in real number format.
     - req_sorter: defines if this solver requires the use of a sorter algorithm. By setting
     it to False, the SolveRoutine will not use the default sorting algorithm.
    """
    real_only: bool = False
    req_sorter: bool = False
    def __init__(self):
        self.own_preconditioner: bool = False

    def __str__(self) -> str:
        return f'{self.__class__.__name__}'
    
    def eig(self, A: lil_matrix, B: np.ndarray, nmodes: int = 6, target_k0: float = None, which: str = 'LM', sign: float = 1.):
        raise NotImplementedError("This classes eigenmdoe solver method is not implemented.")
    
    def reset(self) -> None:
        """Reset state variables like numeric and symbollic factorizations."""
        pass


## -----  SORTERS ----------------------------------------------
@dataclass
class SolveReport:
    solver: str
    sorter: str
    precon: str
    simtime: float
    ndof: int
    nnz: int
    code: int = None


class ReverseCuthillMckee(Sorter):
    """ Implements the Reverse Cuthill-Mckee sorting."""
    def __init__(self):
        super().__init__()
        

    def sort(self, A, b, reuse_sorting: bool = False):
        if not reuse_sorting:
            logger.debug('Generating Reverse Cuthill-Mckee sorting.')
            self.perm = reverse_cuthill_mckee(A)
            self.inv_perm = np.argsort(self.perm)
        logger.debug('Applying Reverse Cuthill-Mckee sorting.')
        Asorted = A[self.perm, :][:, self.perm]
        bsorted = b[self.perm]
        return Asorted, bsorted
    
    def unsort(self, x: np.ndarray):
        logger.debug('Reversing Reverse Cuthill-Mckee sorting.')
        return  x[self.inv_perm]
    
## -----  PRECONS ----------------------------------------------

class ILUPrecon(Preconditioner):
    """ Implements the incomplete LU preconditioner on matrix A. """
    def __init__(self):
        super().__init__()
        self.M = None
        self.fill_factor = 10
        self.options: dict[str, str] = dict(SymmetricMode=True)

    def init(self, A, b):
        logger.info("Generating ILU Preconditioner")
        self.ilu = sparse.linalg.spilu(A, drop_tol=1e-2, fill_factor=self.fill_factor, permc_spec='MMD_AT_PLUS_A', diag_pivot_thresh=0.001, options=self.options)
        self.M = sparse.linalg.LinearOperator(A.shape, self.ilu.solve)

## ----- ITERATIVE SOLVERS -------------------------------------

class SolverBicgstab(Solver):
    """ Implements the Bi-Conjugate Gradient Stabilized method"""
    def __init__(self):
        super().__init__()
        self.atol = 1e-5

        self.A: np.ndarray = None
        self.b: np.ndarray = None

    def callback(self, xk):
        convergence = np.linalg.norm((self.A @ xk - self.b))
        logger.info(f'Iteration {convergence:.4f}')

    def solve(self, A, b, precon, reuse_factorization: bool = False, id: int = -1):
        logger.info(f'Calling BiCGStab. ID={id}')
        self.A = A
        self.b = b
        if precon.M is not None:
            x, info = bicgstab(A, b, M=precon.M, atol=self.atol, callback=self.callback)
        else:
            x, info = bicgstab(A, b, atol=self.atol, callback=self.callback)
        return x, info

class SolverGCROTMK(Solver):
    """ Implements the GCRO-T(m,k) Iterative solver. """
    def __init__(self):
        super().__init__()
        self.atol = 1e-5

        self.A: np.ndarray = None
        self.b: np.ndarray = None

    def callback(self, xk):
        convergence = np.linalg.norm((self.A @ xk - self.b))
        logger.info(f'Iteration {convergence:.4f}')

    def solve(self, A, b, precon, id: int = -1):
        logger.info(f'Calling GCRO-T(m,k) algorithm. ID={id}')
        self.A = A
        self.b = b
        if precon.M is not None:
            x, info = gcrotmk(A, b, M=precon.M, atol=self.atol, callback=self.callback)
        else:
            x, info = gcrotmk(A, b, atol=self.atol, callback=self.callback)
        return x, info

class SolverGMRES(Solver):
    """ Implements the GMRES solver. """
    real_only = False
    req_sorter = True

    def __init__(self):
        super().__init__()
        self.atol = 1e-5

        self.A: np.ndarray = None
        self.b: np.ndarray = None

    def callback(self, norm):
        #convergence = np.linalg.norm((self.A @ xk - self.b))
        logger.info(f'Iteration {norm:.4f}')

    def solve(self, A, b, precon, reuse_factorization: bool = False, id: int = -1):
        logger.info(f'Calling GMRES Function. ID={id}')
        self.A = A
        self.b = b
        if precon.M is not None:
            x, info = gmres(A, b, M=precon.M, atol=self.atol, callback=self.callback, callback_type='pr_norm')
        else:
            x, info = gmres(A, b, atol=self.atol, callback=self.callback, restart=500, callback_type='pr_norm')
        return x, info

## -----  DIRECT SOLVERS ----------------------------------------

class SolverSuperLU(Solver):
    """ Implements Scipi's direct SuperLU solver."""
    req_sorter: bool = False
    real_only: bool = False
    def __init__(self):
        super().__init__()
        self.atol = 1e-5

        self.A: np.ndarray = None
        self.b: np.ndarray = None
        self.options: dict[str, str] = dict(SymmetricMode=True, Equil=False, IterRefine='SINGLE')
        self.lu = None
        
    def solve(self, A, b, precon, reuse_factorization: bool = False, id: int = -1):
        logger.info(f'Calling SuperLU Solver, ID={id}')
        self.single = True
        if not reuse_factorization:
            self.lu = splu(A, permc_spec='MMD_AT_PLUS_A', relax=0, diag_pivot_thresh=0.001, options=self.options)
        x = self.lu.solve(b)

        return x, 0

class SolverUMFPACK(Solver):
    """ Implements the UMFPACK Sparse SP solver."""
    req_sorter = False
    real_only = False

    def __init__(self):
        super().__init__()
        self.A: np.ndarray = None
        self.b: np.ndarray = None
        self.up: um.UmfpackContext = um.UmfpackContext('zi')
        self.up.control[um.UMFPACK_PRL] = 0  #less terminal printing
        self.up.control[um.UMFPACK_IRSTEP] = 2
        self.up.control[um.UMFPACK_STRATEGY] = um.UMFPACK_STRATEGY_SYMMETRIC
        self.up.control[um.UMFPACK_ORDERING] = 3
        self.up.control[um.UMFPACK_PIVOT_TOLERANCE] = 0.001
        self.up.control[um.UMFPACK_SYM_PIVOT_TOLERANCE] = 0.001
        self.up.control[um.UMFPACK_BLOCK_SIZE] = 64
        self.up.control[um.UMFPACK_FIXQ] = -1

        self.fact_symb: bool = False

    def reset(self) -> None:
        self.fact_symb = False

    def solve(self, A, b, precon, reuse_factorization: bool = False, id: int = -1):
        logger.info(f'Calling UMFPACK Solver. ID={id}')
        if self.fact_symb is False:
            logger.debug('Executing symbollic factorization.')
            self.up.symbolic(A)
            #self.up.report_symbolic()
            self.fact_symb = True
        if not reuse_factorization:
            #logger.debug('Executing numeric factorization.')
            self.up.numeric(A)
            self.A = A
        x = self.up.solve(um.UMFPACK_A, self.A, b, autoTranspose = False )
        return x, 0

class SolverPardiso(Solver):
    """ Implements the PARDISO solver through PyPardiso. """
    real_only: bool = True
    req_sorter: bool = False

    def __init__(self):
        super().__init__()

        self.A: np.ndarray = None
        self.b: np.ndarray = None
    
    def solve(self, A, b, precon, reuse_factorization: bool = False, id: int = -1):
        logger.info(f'Calling Pardiso Solver. ID={id}')
        self.A = A
        self.b = b
        try:
            x = pardiso_solve(A, b)
        except PyPardisoError as e:
            print('Error Codes:')
            print(_PARDISO_ERROR_CODES)
        return x, 0
    
## -----  DIRECT EIG SOLVERS --------------------------------------
class SolverLAPACK(EigSolver):

    def __init__(self):
        super().__init__()
    
    def eig(self, 
            A: np.ndarray, 
            B: np.ndarray,
            nmodes: int = 6,
            target_k0: float = 0,
            which: str = 'LM',
            sign: float = 1.0):
        """
        Dense solver for  A x = λ B x   with A = Aᴴ, B = Bᴴ (B may be indefinite).

        Parameters
        ----------
        A, B : (n, n) array_like, complex127/complex64/float64
        k    : int or None
            How many eigenpairs to return.
            * None  → return all n
            * k>0   → return k pairs with |λ| smallest

        Returns
        -------
        lam  : (m,) real ndarray      eigenvalues  (m = n or k)
        vecs : (n, m) complex ndarray eigenvectors, B-orthonormal  (xiᴴ B xj = δij)
        """
        logger.debug('Calling LAPACK eig solver')
        lam, vecs = eig(A.toarray(), B.toarray(), overwrite_a=True, overwrite_b=True, check_finite=False)
        lam, vecs = filter_real_modes(lam, vecs, target_k0, 2, 2, sign=sign)
        return lam, vecs
    
## -----  ITER EIG SOLVERS ---------------------------------------

class SolverARPACK(EigSolver):
    """ Implements the Scipy ARPACK iterative eigenmode solver."""
    def __init__(self):
        super().__init__()

    def eig(self, 
            A: np.ndarray, 
            B: np.ndarray,
            nmodes: int = 6,
            target_k0: float = 0,
            which: str = 'LM',
            sign: float = 1.0):
        logger.info(f'Searching around 	β = {target_k0:.2f} rad/m')
        sigma = sign*(target_k0**2)
        eigen_values, eigen_modes = eigs(A, k=nmodes, M=B, sigma=sigma, which=which)
        return eigen_values, eigen_modes

class SmartARPACK_BMA(EigSolver):
    """ Implements the Scipy ARPACK iterative eigenmode solver with automatic search.
    
    The Solver searches in a geometric range around the target wave constant.
    """
    def __init__(self):
        super().__init__()
        self.symmetric_steps: int = 41
        self.search_range: float = 2.0
        self.energy_limit: float = 1e-4

    def eig(self, 
            A: np.ndarray, 
            B: np.ndarray,
            nmodes: int = 6,
            target_k0: float = 0,
            which: str = 'LM',
            sign: float = 1.):
        logger.info(f'Searching around 	β = {target_k0:.2f} rad/m')
        qs = np.geomspace(1, self.search_range, self.symmetric_steps)
        tot_eigen_values = []
        tot_eigen_modes = []
        energies = []
        for i, q in enumerate(qs):
            # Above target k0
            sigma = sign*((q*target_k0)**2)
            eigen_values, eigen_modes = eigs(A, k=1, M=B, sigma=sigma, which=which)
            energy = np.mean(np.abs(eigen_modes.flatten())**2)
            if energy > self.energy_limit:
                tot_eigen_values.append(eigen_values[0])
                tot_eigen_modes.append(eigen_modes.flatten())
                energies.append(energy)
            if i!=0:
                # Below target k0
                sigma = sign*((target_k0/q)**2)
                eigen_values, eigen_modes = eigs(A, k=1, M=B, sigma=sigma, which=which)
                energy = np.mean(np.abs(eigen_modes.flatten())**2)
                if energy > self.energy_limit:
                    tot_eigen_values.append(eigen_values[0])
                    tot_eigen_modes.append(eigen_modes.flatten())
                    energies.append(energy)
        
        #Sort solutions on mode energy
        val, mode, energy = zip(*sorted(zip(tot_eigen_values,tot_eigen_modes,energies), key=lambda x: x[2], reverse=True))
        eigen_values = np.array(val[:nmodes])
        eigen_modes = np.array(mode[:nmodes]).T

        return eigen_values, eigen_modes
    
class SmartARPACK(EigSolver):
    """ Implements the Scipy ARPACK iterative eigenmode solver with automatic search.
    
    The Solver searches in a geometric range around the target wave constant.
    """
    def __init__(self):
        super().__init__()
        self.symmetric_steps: int = 3
        self.search_range: float = 2.0
        self.energy_limit: float = 1e-4

    def eig(self, 
            A: np.ndarray, 
            B: np.ndarray,
            nmodes: int = 6,
            target_k0: float = 0,
            which: str = 'LM',
            sign: float = 1.):
        logger.info(f'Searching around 	β = {target_k0:.2f} rad/m')
        qs = np.geomspace(1, self.search_range, self.symmetric_steps)
        tot_eigen_values = []
        tot_eigen_modes = []
        for i, q in enumerate(qs):
            # Above target k0
            sigma = sign*((q*target_k0)**2)
            eigen_values, eigen_modes = eigs(A, k=6, M=B, sigma=sigma, which=which)
            for j in range(eigen_values.shape[0]):
                if eigen_values[j]<(sigma/self.search_range):
                    continue
                tot_eigen_values.append(eigen_values[j])
                tot_eigen_modes.append(eigen_modes[:,j])
            if i!=0:
                # Below target k0
                sigma = sign*((target_k0/q)**2)
                eigen_values, eigen_modes = eigs(A, k=6, M=B, sigma=sigma, which=which)
                for j in range(eigen_values.shape[0]):
                    if eigen_values[j]<(sigma/self.search_range):
                        continue
                    tot_eigen_values.append(eigen_values[j])
                    tot_eigen_modes.append(eigen_modes[:,j])
            tot_eigen_values, tot_eigen_modes = filter_unique_eigenpairs(tot_eigen_values, tot_eigen_modes)
            if len(tot_eigen_values)>nmodes:
                break
        #Sort solutions on mode energy
        val, mode = filter_unique_eigenpairs(np.array(tot_eigen_values), np.array(tot_eigen_modes))
        val, mode = zip(*sorted(zip(val,mode), key=lambda x: x[0], reverse=False))
        eigen_values = np.array(val[:nmodes])
        eigen_modes = np.array(mode[:nmodes]).T

        return eigen_values, eigen_modes


## ----- SOLVE ENUMS ---------------------------------------------

class EMSolver(Enum):
    SUPERLU = 1
    UMFPACK = 2
    PARDISO = 3
    LAPACK = 4
    ARPACK = 5
    SMART_ARPACK = 6
    SMART_ARPACK_BMA = 7

    def get_solver(self) -> Solver:
        if self==EMSolver.SUPERLU:
            return SolverSuperLU()
        elif self==EMSolver.UMFPACK:
            if _UMFPACK_AVAILABLE is False:
                return SolverSuperLU()
            else:
                return SolverUMFPACK()
        elif self==EMSolver.PARDISO:
            if _PARDISO_AVAILABLE is False:
                return SolverSuperLU()
            else:
                return SolverPardiso()
        elif self==EMSolver.LAPACK:
            return SolverLAPACK()
        elif self==EMSolver.ARPACK:
            return SolverARPACK()
        elif self==EMSolver.SMART_ARPACK:
            return SmartARPACK()
        elif self==EMSolver.SMART_ARPACK_BMA:
            return SmartARPACK_BMA()
    
## -----  SOLVE ROUTINES -----------------------------------------

class SolveRoutine:
    """ A generic class describing a solve routine.
    A solve routine contains all the relevant sorter preconditioner and solver objects
    and goes through a sequence of steps to solve a linear system or find eigenmodes.

    """
    def __init__(self):
        
        self.sorter: Sorter = ReverseCuthillMckee()
        self.precon: Preconditioner = ILUPrecon()
        self.solvers: dict[EMSolver, Solver] = {slv: slv.get_solver() for slv in EMSolver}

        self.parallel: Literal['SI','MT','MP'] = 'SI'
        self.smart_search: bool = False
        self.forced_solver: list[Solver] = []

        self.use_sorter: bool = False
        self.use_preconditioner: bool = False
        self.use_direct: bool = True

    def __str__(self) -> str:
        return f'SolveRoutine({self.sorter},{self.precon},{self.iterative_solver}, {self.direct_solver})'
    
    def duplicate(self) -> SolveRoutine:
        """Creates a copy of this SolveRoutine class object.

        Returns:
            SolveRoutine: The copied version
        """
        new_routine = self.__class__()
        new_routine.parallel = self.parallel
        new_routine.smart_search = self.smart_search
        new_routine.forced_solver = self.forced_solver
        return new_routine
    
    def set_solver(self, *solver: EMSolver | EigSolver | Solver) -> None:
        """Set a given Solver class instance as the main solver. Solvers will be checked on validity for the given problem.

        Args:
            solver (EMSolver | Solver): The solver objects
        """
        if isinstance(solver, EMSolver):
            self.forced_solver.append(solver.get_solver())
        else:
            self.forced_solver.append(solver)
        
    def configure(self, 
                  parallel: Literal['SI','MT','MP'] = 'SI', smart_search: bool = False) -> SolveRoutine:
        """Configure the solver with the given settings

        Args:
            parallel (Literal['SI','MT','MP'], optional): 
                The solver parallism, Defaults to 'SI'.
                    - "SI" = Single threaded
                    - "MT" = Multi threaded
                    - "MP" = Multi-processing,
            smart_search (bool, optional): Wether to use smart-search solvers for eigenmode problems. Defaults to False.

        Returns:
            SolveRoutine: The same SolveRoutine object.
        """
        self.parallel = parallel
        self.smart_search = smart_search
        return self

    def reset(self) -> None:
        """Reset all solver states"""
        for solver in self.solvers.values():
            solver.reset()
        self.sorter.reset()
        self.parallel: Literal['SI','MT','MP'] = 'SI'
        self.smart_search: bool = False
        self.forced_solver = []

    def _get_solver(self, A: lil_matrix, b: np.ndarray) -> Solver:
        """Returns the relevant Solver object given a certain matrix and source vector

        This is the default implementation for the SolveRoutine Class.
        
        Args:
            A (np.ndarray): The Matrix to solve for
            b (np.ndarray): the vector to solve for

        Returns:
            Solver: Returns the direct solver

        """
        for solver in self.forced_solver:
            if isinstance(solver, Solver):
                return solver
        return self.pick_solver(A,b)
        
    
    def pick_solver(self, A: lil_matrix, b: np.ndarray) -> Solver:
        """Returns the relevant Solver object given a certain matrix and source vector

        This is the default implementation for the SolveRoutine Class.
        
        Args:
            A (np.ndarray): The Matrix to solve for
            b (np.ndarray): the vector to solve for

        Returns:
            Solver: Returns the direct solver

        """
        return self.solvers[EMSolver.SUPERLU]
    
    def _get_eig_solver(self, A: lil_matrix, b: lil_matrix, direct: bool = None) -> Solver:
        """Returns the relevant eigenmode Solver object given a certain matrix and source vector

        This is the default implementation for the SolveRoutine Class.
        
        Args:
            A (np.ndarray): The Matrix to solve for
            b (np.ndarray): the vector to solve for
            direct (bool): If the direct solver should be used.

        Returns:
            Solver: Returns the solver object

        """
        for solver in self.forced_solver:
            if isinstance(solver, EigSolver):
                return solver
        if direct or A.shape[0] < 1000:
            return self.solvers[EMSolver.LAPACK]
        else:
            return self.solvers[EMSolver.SMART_ARPACK]
            
    def _get_eig_solver_bma(self, A: lil_matrix, b: lil_matrix, direct: bool = None) -> Solver:
        """Returns the relevant eigenmode Solver object given a certain matrix and source vector

        This is the default implementation for the SolveRoutine Class.
        
        Args:
            A (np.ndarray): The Matrix to solve for
            b (np.ndarray): the vector to solve for
            direct (bool): If the direct solver should be used.

        Returns:
            Solver: Returns the solver object

        """
        for solver in self.forced_solver:
            if isinstance(solver, EigSolver):
                return solver
        
        if direct or A.shape[0] < 1000:
            return self.solvers[EMSolver.LAPACK]
        else:
            return self.solvers[EMSolver.ARPACK]
    
    def solve(self, A: np.ndarray | lil_matrix | csc_matrix, 
              b: np.ndarray, 
              solve_ids: np.ndarray,
              reuse: bool = False,
              id: int = -1) -> tuple[np.ndarray, SolveReport]:
        """ Solve the system of equations defined by Ax=b for x.

        Solve is the main function call to solve a linear system of equations defined by Ax=b.
        The solve routine will go through the required steps defined in the routine to tackle the problme.

        Args:
            A (np.ndarray | lil_matrix | csc_matrix): The (Sparse) matrix
            b (np.ndarray): The source vector
            solve_ids (np.ndarray): A vector of ids for which to solve the problem. For EM problems this
            implies all non-PEC degrees of freedom.
            reuse (bool): Whether to reuse the existing factorization if it exists.

        Returns:
            np.ndarray: The resultant solution.
        """
        solver = self._get_solver(A, b)

        NF = A.shape[0]
        NS = solve_ids.shape[0]

        A = A.tocsc()

        logger.debug(f'    Removing {NF-NS} prescribed DOFs ({NS} left)')

        Asel = A[np.ix_(solve_ids, solve_ids)]
        bsel = b[solve_ids]

        if solver.real_only:
            logger.debug('    Converting to real matrix')
            Asel, bsel = complex_to_real_block(Asel, bsel)

        # SORT
        sorter = 'None'
        if solver.req_sorter and self.use_sorter:
            sorter = str(self.sorter)
            Asorted, bsorted = self.sorter.sort(Asel, bsel, reuse_sorting=reuse)
        else:
            Asorted, bsorted = Asel, bsel
        
        # Preconditioner
        precon = 'None'
        if self.use_preconditioner:
            if not self.iterative_solver.own_preconditioner:
                self.precon.init(Asorted, bsorted)
                precon = str(self.precon)

        start = time.time()
        x_solved, code = solver.solve(Asorted, bsorted, self.precon, reuse_factorization=reuse, id=id)
        end = time.time()
        simtime = end-start
        logger.info(f'Time taken: {simtime:.3f} seconds')
        logger.debug(f'    O(N²) performance = {(NS**2)/((end-start+1e-6)*1e6):.3f} MDoF/s')
        
        if self.use_sorter and solver.req_sorter:
            x = self.sorter.unsort(x_solved)
        else:
            x = x_solved
        
        if solver.real_only:
            logger.debug('    Converting back to complex matrix')
            x = real_to_complex_block(x)

        solution = np.zeros((NF,), dtype=np.complex128)
        
        solution[solve_ids] = x

        logger.debug('Solver complete!')
        if code:
            logger.debug('    Solver code: {code}')
        return solution, SolveReport(str(solver), sorter, precon, simtime, NS, A.nnz, code)
    
    def eig_boundary(self, 
            A: np.ndarray | lil_matrix | csc_matrix, 
            B: np.ndarray, 
            solve_ids: np.ndarray,
            nmodes: int = 6,
            direct: bool = None,
            target_k0: float = None,
            which: str = 'LM', 
            sign: float=-1) -> tuple[np.ndarray, np.ndarray, SolveReport]:
        """ Find the eigenmodes for the system Ax = λBx for a boundary mode problem

        For generalized eigenvalue problems of boundary mode analysis studies, the equation is: Ae = -β²Be

        Args:
            A (csc_matrix): The Stiffness matrix
            B (csc_matrix): The mass matrix
            solve_ids (np.ndarray): The free nodes (non PEC)
            nmodes (int): The number of modes to solve for. Defaults to 6
            direct (bool): If the direct solver should be used (always). Defaults to False
            target_k0 (float): The k0 value to search around
            which (str): The search method. Defaults to 'LM' (Largest Magnitude)
            sign (float): The sign of the eigenvalue expression. Defaults to -1

        Returns:
            np.ndarray: The eigen values
            np.ndarray: The eigen vectors
            SolveReport: The solution report
        """
        solver = self._get_eig_solver_bma(A, B, direct=direct)

        NF = A.shape[0]
        NS = solve_ids.shape[0]

        logger.debug(f'    Removing {NF-NS} prescribed DOFs ({NS} left)')

        Asel = A[np.ix_(solve_ids, solve_ids)]
        Bsel = B[np.ix_(solve_ids, solve_ids)]
        
        start = time.time()
        eigen_values, eigen_modes = solver.eig(Asel, Bsel, nmodes, target_k0, which, sign=sign)
        end = time.time()

        simtime = end-start
        return eigen_values, eigen_modes, SolveReport(str(solver), 'None', 'None', simtime, NS, A.nnz, simtime)

    def eig(self, 
            A: np.ndarray | lil_matrix | csc_matrix, 
            B: np.ndarray, 
            solve_ids: np.ndarray,
            nmodes: int = 6,
            direct: bool = None,
            target_f0: float = None,
            which: str = 'LM') -> tuple[np.ndarray, np.ndarray, SolveReport]:
        """
        Find the eigenmodes for the system Ax = λBx for a boundary mode problem
        
        Args:
            A (csc_matrix): The Stiffness matrix
            B (csc_matrix): The mass matrix
            solve_ids (np.ndarray): The free nodes (non PEC)
            nmodes (int): The number of modes to solve for. Defaults to 6
            direct (bool): If the direct solver should be used (always). Defaults to False
            target_k0 (float): The k0 value to search around
            which (str): The search method. Defaults to 'LM' (Largest Magnitude)
            sign (float): The sign of the eigenvalue expression. Defaults to -1\
        Returns:
            np.ndarray: The resultant solution.
        """
        solver = self._get_eig_solver(A, B, direct=direct)

        NF = A.shape[0]
        NS = solve_ids.shape[0]

        logger.debug(f'    Removing {NF-NS} prescribed DOFs ({NS} left)')

        Asel = A[np.ix_(solve_ids, solve_ids)]
        Bsel = B[np.ix_(solve_ids, solve_ids)]
        
        start = time.time()
        eigen_values, eigen_modes = solver.eig(Asel, Bsel, nmodes, target_f0, which, sign=1.0)
        end = time.time()
        simtime = end-start

        Nsols = eigen_modes.shape[1]
        sols = np.zeros((NF, Nsols), dtype=np.complex128)
        for i in range(Nsols):
            sols[solve_ids,i] = eigen_modes[:,i]

        return eigen_values, sols, SolveReport(str(solver), 'None', 'None', simtime, NS, A.nnz, simtime)


class AutomaticRoutine(SolveRoutine):
    """ Defines the Automatic Routine for EMerge.
    """
        
    def pick_solver(self, A: np.ndarray, b: np.ndarray) -> Solver:
        """Returns the relevant Solver object given a certain matrix and source vector

        The current implementation only looks at matrix size to select the best solver. Matrices 
        with a large size will use iterative solvers while smaller sizes will use either Pardiso
        for medium sized problems or SPSolve for small ones.
        
        Args:
            A (np.ndarray): The Matrix to solve for
            b (np.ndarray): the vector to solve for

        Returns:
            Solver: A solver object appropriate for solving the problem.

        """
        N = b.shape[0]
        if N < 10_000:
            return self.solvers[EMSolver.SUPERLU]
        if self.parallel=='SI':
            if _PARDISO_AVAILABLE:
                return self.solvers[EMSolver.PARDISO]
            elif _UMFPACK_AVAILABLE:
                return self.solvers[EMSolver.UMFPACK]
            else:
                return self.solvers[EMSolver.SUPERLU]
        elif self.parallel=='MP':
            if _UMFPACK_AVAILABLE:
                return self.solvers[EMSolver.UMFPACK]
            else:
                return self.solvers[EMSolver.SUPERLU]
        elif self.parallel=='MT':
            return self.solvers[EMSolver.SUPERLU]
        return self.solvers[EMSolver.SUPERLU]
    
DEFAULT_ROUTINE = AutomaticRoutine()


from _emerge.material import Material, AIR, COPPER

C0 = 299792458
Z0 = 376.73031366857
PI = 3.14159265358979323846
EISO: float = (Z0/(2*PI))**0.5
EOMNI = (3*Z0/(4*PI))**0.5
EPS0 = 8.854187818814e-12
MU0 = 1.2566370612720e-6

### MATERIALS 
VACUUM = Material(color="#2d8cd5", opacity=0.05)
FR4 = Material(er=4.4, tand=0.001, color="#3c9747")
# Conductors
ALUMINUM  = Material(cond=3.5e7,         color="#aaaaaa")
GOLD      = Material(cond=4.1e7,         color="#A07130")
SILVER    = Material(cond=6.3e7,         color="#bfbfbf")
TIN       = Material(cond=9.17e6,        color="#d6d6d6")
NICKEL    = Material(cond=1.43e7,        color="#78706c")
IRON      = Material(cond=1.0e7,  ur=5000, color="#666666")
STEEL     = Material(cond=1.45e6, ur=100,  color="#808080")

# Semiconductors & Dielectrics
SILICON   = Material(er=11.68,    cond=0.1, color="#333333")
SIO2      = Material(er=3.9,                color="#e6e6e6")
GAAS      = Material(er=12.9,    cond=0.0, color="#404071")

# Dielectrics
PTFE      = Material(er=2.1,     tand=0.0002,       color="#cccccc")
POLYIMIDE = Material(er=3.4,     tand=0.02,         color="#b8b8b8")
CERAMIC   = Material(er=6.0,     tand=0.001,        color="#efead1")

# Liquids
WATER     = Material(er=80.1,    cond=0.0,  color="#0080ff", opacity=0.3)
FERRITE   = Material(er=12.0,    ur=2000, tand=0.02, color="#994d4d")

# Specialty RF Substrates
ROGERS_4350B = Material(er=3.66, tand=0.0037,       color="#3c9747")
ROGERS_5880  = Material(er=2.2,  tand=0.0009,       color="#3c9747")
ROGERS_RO3003   = Material(er=3.0,  tand=0.0013,      color="#3c9747")
ROGERS_RO3010   = Material(er=10.2, tand=0.0023,      color="#3c9747")
ROGERS_RO4003C  = Material(er=3.55, tand=0.0027,      color="#3c9747")
ROGERS_DUROID6002 = Material(er=2.94, tand=0.0012,    color="#3c9747")
ROGERS_DUROID6010 = Material(er=6.15, tand=0.0027,    color="#3c9747")
ROGERS_RT5880  = Material(er=2.2,  tand=0.0009,       color="#3c9747")
TACONIC_RF35   = Material(er=3.5,  tand=0.0018,       color="#3c9747")
TACONIC_TLC30  = Material(er=3.0,  tand=0.0020,       color="#3c9747")
ISOLA_I_TERA_MT = Material(er=3.45, tand=0.0030,      color="#3c9747")
ISOLA_NELCO_4000_13 = Material(er=3.77, tand=0.008,  color="#3c9747")
VENTEC_VERDELAY_400HR = Material(er=4.0,  tand=0.02,  color="#3c9747")

# Legacy FR Materials
FR1            = Material(er=4.8,  tand=0.025,        color="#3c9747")
FR2            = Material(er=4.8,  tand=0.02,        color="#3c9747")

# Magnetic Materials
MU_METAL  = Material(cond=1.0e6, ur=200000, color="#666680")


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Andrés Martínez Mera.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

import os
from typing import Literal
import numpy as np
from datetime import datetime
from loguru import logger

_F_UNIT = {'HZ': 1,
           'KHZ': 1e3,
           'MHZ': 1e6,
           'GHZ': 1e9,
}
def generate_touchstone(filename: str,
                        freq: np.ndarray,
                        Smat: np.ndarray,
                        data_format: Literal['RI','MA','DB'],
                        custom_comments: list[str] = None,
                        funit: Literal['HZ','KHZ','MHZ','GHZ'] = 'GHZ') -> None:
    """
    Export S-parameter data to a Touchstone file

    Parameters
    ----------
    filename : str
        Base filename (may include path). If no extension is given, one
        will be added of the form '.sNp' where N = number of ports.
    freq : np.ndarray
        1-D array of M frequency points, in Hz.
    Smat : np.ndarray
        3-D array of S-parameters with shape (M, N, N).
    data_format : {'RI','MA','DB'}
        RI = real&imag, MA = magnitude&angle (deg), DB = dB&angle (deg).
    custom_comments : list[str], optional
        List of custom comment strings to add to the touchstone file header.
        Each string will be prefixed with "! " automatically.
    """
    # --- validations ---
    if Smat.ndim != 3:
        raise ValueError(f"Smat must be 3-D with shape (M,N,N), got ndim={Smat.ndim}")
    M, Nports, N2 = Smat.shape
    if Nports != N2:
        raise ValueError(f"Smat must be square in its last two dims, got {Nports}×{N2}")
    if freq.ndim != 1 or freq.size != M:
        raise ValueError(f"freq must be 1-D of length {M}, got shape {freq.shape}")

    # --- determine output filename & extension ---
    base, ext = os.path.splitext(filename)
    if ext == '':
        ext = f".s{Nports}p"
    filename_out = base + ext

    # --- write the Touchstone file ---
    with open(filename_out, 'w') as f:
        # Write header
        f.write(f"! Generated by EMerge - {Nports}-port S-parameters\n")
        timestamp_str = datetime.now().strftime("%Y-%m-%d %H:%M %Z")
        f.write(f"! File generated: {timestamp_str}\n")
        
        # Write custom comments if provided
        if custom_comments:
            for comment in custom_comments:
                # Ensure comment starts with "! " if not already present
                if not comment.startswith("! "):
                    comment = "! " + comment
                f.write(f"{comment}\n")

        f.write(f"# {funit} S {data_format.upper()} R 50\n")

        # Write data
        for i in range(M):
            freq_ghz = freq[i] / _F_UNIT[funit]  # Convert Hz to GHz
            
            # Build all S-parameter values for this frequency
            s_values = []
            for row in range(Nports):
                for col in range(Nports):
                    s_val = Smat[i, row, col]
                    
                    if data_format == 'RI':
                        # Real and Imaginary
                        val1 = s_val.real
                        val2 = s_val.imag
                    elif data_format == 'MA':
                        # Magnitude and Angle (degrees)
                        val1 = np.abs(s_val)
                        val2 = np.angle(s_val) * 180.0 / np.pi
                    elif data_format == 'DB':
                        # dB and Angle (degrees)
                        val1 = 20.0 * np.log10(np.abs(s_val)) if np.abs(s_val) > 0 else -300.0
                        val2 = np.angle(s_val) * 180.0 / np.pi
                    
                    s_values.extend([val1, val2])
            
            # Write frequency and S-parameters according to touchstone format
            if Nports == 1:
                # s1p: freq S11_real S11_imag (3 values per line)
                f.write(f"{freq_ghz:12.6e}  {s_values[0]:12.6e}  {s_values[1]:12.6e}\n")
            
            elif Nports == 2:
                # s2p: freq S11_r S11_i S21_r S21_i S12_r S12_i S22_r S22_i (9 values per line)
                f.write(f"{freq_ghz:12.6e}")
                for val in s_values:
                    f.write(f"  {val:12.6e}")
                f.write("\n")
            
            else:
                # s3p, s4p, etc: freq on first line, then S-parameters on subsequent lines
                f.write(f"{freq_ghz:12.6e}")
                
                if (Nports == 3):
                    values_per_line = 6  # 3-port S-parameter
                else: 
                    values_per_line = 8  # 4-port S-parameter pairs
                for j in range(0, len(s_values), values_per_line):
                    f.write("   ")  # Indent continuation lines
                    if (j != 0):
                        f.write(' ' * 12)  # Extra indentation
                    end_idx = min(j + values_per_line, len(s_values))
                    for val in s_values[j:end_idx]:
                        f.write(f" {val:12.6e}")
                    f.write("\n")
    
    logger.info(f"Touchstone file written to '{filename_out}'")


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

import gmsh
from ..geometry import GeoObject, GeoSurface, GeoVolume
from ..cs import CoordinateSystem, GCS
import numpy as np
from enum import Enum
from .operations import subtract
from ..selection import FaceSelection, Selector, SELECTOR_OBJ

from typing import Literal
from functools import reduce

class Alignment(Enum):
    CENTER = 1
    CORNER = 2

class Box(GeoVolume):
    """ A class that represents a box shaped volume

    """

    def __init__(self, 
                 width: float, 
                 depth: float, 
                 height: float, 
                 position: tuple = (0,0,0),
                 alignment: Alignment = Alignment.CORNER,
                 cs: CoordinateSystem = GCS):
        """Creates a box volume object.
        Specify the alignment of the box with the provided position. The options are CORNER (default)
        for the front-left-bottom node of the box or CENTER for the center of the box.

        Args:
            width (float): The x-size
            depth (float): The y-size
            height (float): The z-size
            position (tuple, optional): The position of the box. Defaults to (0,0,0).
            alignment (Alignment, optional): Which point of the box is placed at the position. 
                Defaults to Alignment.CORNER.
        """
        if alignment is Alignment.CENTER:
            position = (position[0]-width/2, position[1]-depth/2, position[2]-height/2)
        
        x,y,z = position

        tag = gmsh.model.occ.addBox(x,y,z,width,depth,height)
        super().__init__(tag)

        self.center = (x+width/2, y+depth/2, z+height/2)
        self.width = width
        self.height = height
        self.depth = depth

        wax = cs.xax.np
        dax = cs.yax.np
        hax = cs.zax.np

        p0 = self.center
        pc = p0
        self._add_face_pointer('front', pc - depth/2*dax, -dax)
        self._add_face_pointer('back', pc + depth/2*dax, dax)
        self._add_face_pointer('left', pc - width/2*wax, -wax)
        self._add_face_pointer('right', pc + width/2*wax, wax)
        self._add_face_pointer('top', pc + height/2*hax, hax)
        self._add_face_pointer('bottom', pc - height/2*hax, -hax)
        
    
    def outside(self, *exclude: Literal['bottom','top','right','left','front','back']) -> FaceSelection:
        """Select all outside faces except for the once specified by outside

        Returns:
            FaceSelection: The resultant face selection
        """
        tagslist = [self._face_tags(name) for name in  ['bottom','top','right','left','front','back'] if name not in exclude]
        
        tags = list(reduce(lambda a,b: a+b, tagslist))
        return FaceSelection(tags)

    
class Sphere(GeoVolume):

    def __init__(self, 
                 radius: float,
                 position: tuple = (0,0,0)):
        """Generates a sphere objected centered ont he position with the given radius

        Args:
            radius (float): The sphere radius
            position (tuple, optional): The center position. Defaults to (0,0,0).
        """
        super().__init__([])
        x,y,z = position
        self.tags: list[int] = [gmsh.model.occ.addSphere(x,y,z,radius),]

class XYPlate(GeoSurface):
    def __init__(self, 
                 width: float, 
                 depth: float, 
                 position: tuple = (0,0,0),
                 alignment: Alignment = Alignment.CORNER):
        """Generates and XY-plane oriented plate
        
        Specify the alignment of the plate with the provided position. The options are CORNER (default)
        for the front-left node of the plate or CENTER for the center of the plate.

        Args:
            width (float): The x-size of the plate
            depth (float): The y-size of the plate
            position (tuple, optional): The position of the alignment node. Defaults to (0,0,0).
            alignment (Alignment, optional): Which node to align to. Defaults to Alignment.CORNER.
        """
        super().__init__([])
        if alignment is Alignment.CENTER:
            position = (position[0]-width/2, position[1]-depth/2, position[2])
        
        x,y,z = position
        self.tags: list[int] = [gmsh.model.occ.addRectangle(x,y,z,width,depth),]


class Plate(GeoSurface):
        
    def __init__(self,
                origin: tuple[float, float, float],
                u: tuple[float, float, float],
                v: tuple[float, float, float]):
        """A generalized 2D rectangular plate in XYZ-space.

        The plate is specified by an origin (o) in meters coordinate plus two vectors (u,v) in meters
        that span two of the sides such that all points of the plate are defined by:
            p1 = o
            p2 = o+u
            p3 = o+v
            p4 = o+u+v
        Args:
            origin (tuple[float, float, float]): The origin of the plate in meters
            u (tuple[float, float, float]): The u-axis of the plate
            v (tuple[float, float, float]): The v-axis of the plate
        """
        
        origin = np.array(origin)
        u = np.array(u)
        v = np.array(v)
        
        tagp1 = gmsh.model.occ.addPoint(*origin)
        tagp2 = gmsh.model.occ.addPoint(*(origin+u))
        tagp3 = gmsh.model.occ.addPoint(*(origin+v))
        tagp4 = gmsh.model.occ.addPoint(*(origin+u+v))

        tagl1 = gmsh.model.occ.addLine(tagp1, tagp2)
        tagl2 = gmsh.model.occ.addLine(tagp2, tagp4)
        tagl3 = gmsh.model.occ.addLine(tagp4, tagp3)
        tagl4 = gmsh.model.occ.addLine(tagp3, tagp1)

        tag_wire = gmsh.model.occ.addWire([tagl1,tagl2, tagl3, tagl4])

        tags: list[int] = [gmsh.model.occ.addPlaneSurface([tag_wire,]),]
        super().__init__(tags)

class Cyllinder(GeoVolume):

    def __init__(self, 
                 radius: float,
                 height: float,
                 cs: CoordinateSystem = None,
                 Nsections: int = None):
        """Generates a Cyllinder object in 3D space.
        The cyllinder will always be placed in the origin of the provided CoordinateSystem.
        The bottom cyllinder plane is always placed in the XY-plane. The lenth of the cyllinder is
        oriented along the Z-axis.

        By default the cyllinder uses the Open Cascade modeling for a cyllinder. In this representation
        the surface of the cyllinder is approximated with a tolerance thay may be irregular.
        As an alternative, the argument Nsections may be provided in which case the Cyllinder is replaced
        by an extrusion of a regular N-sided polygon.

        Args:
            radius (float): The radius of the Cyllinder
            height (float): The height of the Cyllinder
            cs (CoordinateSystem, optional): The coordinate system. Defaults to None.
            Nsections (int, optional): The number of sections. Defaults to None.
        """
        ax = cs.zax.np

        if Nsections:
            from .polybased import XYPolygon
            cyl = XYPolygon.circle(radius, Nsections=Nsections).extrude(height, cs)
            cyl._exists = False
            self._face_pointers = cyl._face_pointers
            super().__init__(cyl.tags)
        else:
            cyl = gmsh.model.occ.addCylinder(cs.origin[0], cs.origin[1], cs.origin[2],
                                         height*ax[0], height*ax[1], height*ax[2],
                                         radius)
            super().__init__(cyl)
        self._add_face_pointer('front', cs.origin, -cs.zax.np)
        self._add_face_pointer('back', cs.origin+height*cs.zax.np, cs.zax.np)
            
        self.cs: CoordinateSystem = cs
        self.radius = radius
        self.height = height

    def face_points(self, nRadius: int = 10, Angle: int = 10, face_number: int = 1) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Returns the points of the cylinder."""
        rs = np.linspace(0, self.radius, nRadius)
        angles = np.linspace(0, 2 * np.pi, int(360 / Angle), endpoint=False)
        R, A = np.meshgrid(rs, angles)
        x = R * np.cos(A)
        y = R * np.sin(A)
        z = np.zeros_like(x)
        if face_number == 2:
            z = z + self.height  

        xo, yo, zo = self.cs.in_global_cs(x.flatten(), y.flatten(), z.flatten())
        return xo, yo, zo

class CoaxCyllinder(GeoVolume):
    """A coaxial cylinder with an inner and outer radius."""
    
    def __init__(self, 
                 rout: float,
                 rin: float,
                 height: float,
                 cs: CoordinateSystem = None,
                 Nsections: int = None):
        """Generates a Coaxial cyllinder object in 3D space.
        The coaxial cyllinder will always be placed in the origin of the provided CoordinateSystem.
        The bottom coax plane is always placed in the XY-plane. The lenth of the coax is
        oriented along the Z-axis.

        By default the coax uses the Open Cascade modeling for a cyllinder. In this representation
        the surface of the cyllinder is approximated with a tolerance thay may be irregular.
        As an alternative, the argument Nsections may be provided in which case the Cyllinder is replaced
        by an extrusion of a regular N-sided polygon.

        Args:
            radius (float): The radius of the Cyllinder
            height (float): The height of the Cyllinder
            cs (CoordinateSystem, optional): The coordinate system. Defaults to None.
            Nsections (int, optional): The number of sections. Defaults to None.
        """
        if cs is None:
            cs = GCS
        if rout <= rin:
            raise ValueError("Outer radius must be greater than inner radius.")
        
        self.rout = rout
        self.rin = rin
        self.height = height

        self.cyl_out = Cyllinder(rout, height, cs, Nsections=Nsections)
        self.cyl_in = Cyllinder(rin, height, cs, Nsections=Nsections)
        self.cyl_in._exists = False
        self.cyl_out._exists = False
        cyltags, _ = gmsh.model.occ.cut(self.cyl_out.dimtags, self.cyl_in.dimtags)
        
        super().__init__([dt[1] for dt in cyltags])

        self._add_face_pointer('front', cs.origin, -cs.zax.np)
        self._add_face_pointer('back', cs.origin+height*cs.zax.np, cs.zax.np)

        self.cs = cs

    def face_points(self, nRadius: int = 10, Angle: int = 10, face_number: int = 1) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Returns the points of the coaxial cylinder."""
        rs = np.linspace(self.rin, self.rout, nRadius)
        angles = np.linspace(0, 2 * np.pi, int(360 / Angle), endpoint=False)
        R, A = np.meshgrid(rs, angles)
        x = R * np.cos(A)
        y = R * np.sin(A)
        z = np.zeros_like(x)
        if face_number == 2:
            z = z + self.height  

        xo, yo, zo = self.cs.in_global_cs(x.flatten(), y.flatten(), z.flatten())
        return xo, yo, zo
        return super().boundary()
    
class HalfSphere(GeoVolume):

    def __init__(self,
                 radius: float,
                 position: tuple = (0,0,0),
                 direction: tuple = (1,0,0)):
        super().__init__([])
        sphere = Sphere(radius, position=position)
        cx, cy, cz = position

        box = Box(1.1*radius, 2.2*radius, 2.2*radius, position=(cx-radius*1.1,cy-radius*1.1, cz-radius*1.1))

        self.tag = subtract(sphere, box)[0].tag



class OldBox(GeoVolume):
    '''The sided box class creates a box just like the Box class but with selectable face tags.
    This class is more convenient in use when defining radiation boundaries.'''
    def __init__(self, 
                 width: float, 
                 depth: float, 
                 height: float, 
                 position: tuple = (0,0,0),
                 cs: CoordinateSystem = GCS,
                 alignment: Alignment = Alignment.CORNER):
        """Creates a box volume object with selectable sides.

        Specify the alignment of the box with the provided position. The options are CORNER (default)
        for the front-left-bottom node of the box or CENTER for the center of the box.

        Args:
            width (float): The x-size
            depth (float): The y-size
            height (float): The z-size
            position (tuple, optional): The position of the box. Defaults to (0,0,0).
            alignment (Alignment, optional): Which point of the box is placed at the position. 
                Defaults to Alignment.CORNER.
        """
        

        if alignment is Alignment.CORNER:
            position = (position[0]+width/2, position[1]+depth/2, position[2])
        elif alignment is Alignment.CENTER:
            position = (position[0], position[1], position[2] - height/2)
        p0 = np.array(position)
        p1 = p0 + cs.zax.np * height

        wax = cs.xax.np
        dax = cs.yax.np
        hax = cs.zax.np

        w = width
        d = depth

        p11 = p0 + wax * w/2 + dax * d/2 # right back
        p12 = p0 + wax * w/2 - dax * d/2 # right front
        p13 = p0 - wax * w/2 - dax * d/2 # Left front
        p14 = p0 - wax * w/2 + dax * d/2 # left back

        p21 = p1 + wax * w/2 + dax * d/2
        p22 = p1 + wax * w/2 - dax * d/2
        p23 = p1 - wax * w/2 - dax * d/2
        p24 = p1 - wax * w/2 + dax * d/2

        pt11 = gmsh.model.occ.addPoint(*p11) # right back
        pt12 = gmsh.model.occ.addPoint(*p12) # Right front
        pt13 = gmsh.model.occ.addPoint(*p13) # left front
        pt14 = gmsh.model.occ.addPoint(*p14) # Left back
        pt21 = gmsh.model.occ.addPoint(*p21)
        pt22 = gmsh.model.occ.addPoint(*p22)
        pt23 = gmsh.model.occ.addPoint(*p23)
        pt24 = gmsh.model.occ.addPoint(*p24)

        l1r = gmsh.model.occ.addLine(pt11, pt12) #Right
        l1f = gmsh.model.occ.addLine(pt12, pt13) #Front
        l1l = gmsh.model.occ.addLine(pt13, pt14) #Left
        l1b = gmsh.model.occ.addLine(pt14, pt11) #Back
        
        l2r = gmsh.model.occ.addLine(pt21, pt22)
        l2f = gmsh.model.occ.addLine(pt22, pt23)
        l2l = gmsh.model.occ.addLine(pt23, pt24)
        l2b = gmsh.model.occ.addLine(pt24, pt21)

        dbr = gmsh.model.occ.addLine(pt11, pt21)
        dfr = gmsh.model.occ.addLine(pt12, pt22)
        dfl = gmsh.model.occ.addLine(pt13, pt23)
        dbl = gmsh.model.occ.addLine(pt14, pt24)

        wbot = gmsh.model.occ.addWire([l1r, l1b, l1l, l1f])
        wtop = gmsh.model.occ.addWire([l2r, l2b, l2l, l2f])
        wright = gmsh.model.occ.addWire([l1r, dbr, l2r, dfr])
        wleft = gmsh.model.occ.addWire([l1l, dbl, l2l, dfl])
        wback = gmsh.model.occ.addWire([l1b, dbl, l2b, dbr])
        wfront = gmsh.model.occ.addWire([l1f, dfr, l2f, dfl])

        bottom_tag = gmsh.model.occ.addSurfaceFilling(wbot)
        top_tag = gmsh.model.occ.addSurfaceFilling(wtop)
        front_tag = gmsh.model.occ.addSurfaceFilling(wfront)
        back_tag = gmsh.model.occ.addSurfaceFilling(wback)
        left_tag = gmsh.model.occ.addSurfaceFilling(wleft)
        right_tag = gmsh.model.occ.addSurfaceFilling(wright)

        sv = gmsh.model.occ.addSurfaceLoop([bottom_tag,
                                            top_tag,
                                            right_tag,
                                            left_tag,
                                            front_tag,
                                            back_tag])

        volume_tag: int = gmsh.model.occ.addVolume([sv,])

        super().__init__(volume_tag)
        #self.tags: list[int] = [volume_tag,]

        pc = p0 + height/2*hax
        self._add_face_pointer('front', pc - depth/2*dax, -dax)
        self._add_face_pointer('back', pc + depth/2*dax, dax)
        self._add_face_pointer('left', pc - width/2*wax, -wax)
        self._add_face_pointer('right', pc + width/2*wax, wax)
        self._add_face_pointer('top', pc + height/2*hax, hax)
        self._add_face_pointer('bottom', pc - height/2*hax, -hax)
        
    
    def outside(self, *exclude: Literal['bottom','top','right','left','front','back']) -> FaceSelection:
        """Select all outside faces except for the once specified by outside

        Returns:
            FaceSelection: The resultant face selection
        """
        tagslist = [self._face_tags(name) for name in  ['bottom','top','right','left','front','back'] if name not in exclude]
        
        tags = list(reduce(lambda a,b: a+b, tagslist))
        return FaceSelection(tags)



# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

import numpy as np
from ..microwave_bc import PEC, BoundaryCondition, RectangularWaveguide, RobinBC, PortBC, Periodic
from ....elements.nedelec2 import Nedelec2
from ....elements.nedleg2 import NedelecLegrange2
from ....mth.optimized import gaus_quad_tri
from scipy.sparse import csr_matrix, eye
from loguru import logger
from ..simjob import SimJob
from collections import defaultdict

C0 = 299792458
EPS0 = 8.854187818814e-12

def matprint(mat: np.ndarray) -> None:
    factor = np.max(np.abs(mat.flatten()))
    if factor == 0:
        factor = 1
    print(mat.real/factor)

def select_bc(bcs: list[BoundaryCondition], bctype):
    return [bc for bc in bcs if isinstance(bc,bctype)]

def diagnose_matrix(mat: np.ndarray) -> None:

    if not isinstance(mat, np.ndarray):
        logger.debug('Converting sparse array to flattened array')
        mat = mat[mat.nonzero()].A1
        #mat = np.array(nonzero_mat)
    
    ''' Prints all indices of Nan's and infinities in a matrix '''
    ids = np.where(np.isnan(mat))
    if len(ids[0]) > 0:
        logger.error(f'Found NaN at {ids}')
    ids = np.where(np.isinf(mat))
    if len(ids[0]) > 0:
        logger.error(f'Found Inf at {ids}')
    ids = np.where(np.abs(mat) > 1e10)
    if len(ids[0]) > 0:
        logger.error(f'Found large values at {ids}')
    logger.info('Diagnostics finished')

#############
def gen_key(coord, mult):
    return tuple([int(round(c*mult)) for c in coord])

def plane_basis_from_points(points: np.ndarray) -> np.ndarray:
    """
    Compute an orthonormal basis from a cloud of 3D points dominantly
    lying on one plane.

    Parameters
    ----------
    points : ndarray, shape (3, N)
        3D coordinates of the point cloud.

    Returns
    -------
    basis : ndarray, shape (3, 3)
        Matrix whose columns are:
            - first principal direction (plane X axis)
            - second principal direction (plane Y axis)
            - plane normal vector (Z axis)
    """
    if points.shape[0] != 3:
        raise ValueError("Input must have shape (3, N)")

    # Compute centroid
    centroid = points.mean(axis=1, keepdims=True)

    # Center the data
    points_centered = points - centroid

    # Compute covariance matrix (3x3)
    C = (points_centered @ points_centered.T) / points.shape[1]

    # Eigen decomposition
    eigvals, eigvecs = np.linalg.eigh(C)

    # Sort eigenvectors by descending eigenvalue
    idx = np.argsort(eigvals)[::-1]
    eigvecs = eigvecs[:, idx]

    # Columns of eigvecs = principal axes
    return eigvecs

class Assembler:

    def __init__(self):
        self.cached_matrices = None
        self.conductivity_limit = 1e7

    
    def assemble_bma_matrices(self,
                              field: Nedelec2,
                        er: np.ndarray, 
                        ur: np.ndarray, 
                        sig: np.ndarray,
                        k0: float,
                        port: PortBC,
                        bcs: list[BoundaryCondition]) -> tuple[np.ndarray, np.ndarray, np.ndarray, NedelecLegrange2]:
        
        from .generalized_eigen import generelized_eigenvalue_matrix
        
        mesh = field.mesh
        tri_ids = mesh.get_triangles(port.tags)

        origin = tuple([c-n for c,n in zip(port.cs.origin, port.cs.gzhat)])

        boundary_surface = mesh.boundary_surface(port.tags, origin)
        nedlegfield = NedelecLegrange2(boundary_surface, port.cs)

        ermesh = er[:,:,tri_ids]
        urmesh = ur[:,:,tri_ids]
        sigmesh = sig[tri_ids]

        ermesh = ermesh - 1j * sigmesh/(k0*C0*EPS0)

        E, B = generelized_eigenvalue_matrix(nedlegfield, ermesh, urmesh, port.cs._basis, k0)

        pecs: list[PEC] = [bc for bc in bcs if isinstance(bc,PEC)]
        if len(pecs) > 0:
            logger.debug('Implementing PEC BCs')

        pec_ids = []

        # Process all concutors. Everything above the conductivity limit is considered pec.
        for it in range(boundary_surface.n_tris):
            if sigmesh[it] > self.conductivity_limit:
                pec_ids.extend(list(nedlegfield.tri_to_field[:,it]))

        # Process all PEC Boundary Conditions
        pec_edges = []
        pec_vertices = []
        for pec in pecs:
            face_tags = pec.tags

            tri_ids = mesh.get_triangles(face_tags)
            edge_ids = list(mesh.tri_to_edge[:,tri_ids].flatten())
            for ii in edge_ids:
                i2 = nedlegfield.mesh.from_source_edge(ii)
                if i2 is None:
                    continue
                eids = nedlegfield.edge_to_field[:, i2]
                pec_ids.extend(list(eids))
                pec_edges.append(eids[0])
                pec_vertices.append(eids[3]-nedlegfield.n_xy)
                pec_vertices.append(eids[4]-nedlegfield.n_xy)
            

            for ii in tri_ids:
                i2 = nedlegfield.mesh.from_source_tri(ii)
                if i2 is None:
                    continue
                tids = nedlegfield.tri_to_field[:, i2]
                pec_ids.extend(list(tids))

        port._field = nedlegfield
        port._pece = pec_edges
        port._pecv = pec_vertices
        # Process all port boundary Conditions
        pec_ids = set(pec_ids)
        solve_ids = [i for i in range(nedlegfield.n_field) if i not in pec_ids]

        return E, B, np.array(solve_ids), nedlegfield

    def assemble_freq_matrix(self, field: Nedelec2, 
                        er: np.ndarray, 
                        ur: np.ndarray, 
                        sig: np.ndarray,
                        bcs: list[BoundaryCondition],
                        frequency: float,
                        cache_matrices: bool = False) -> SimJob:
        
        from .curlcurl import tet_mass_stiffness_matrices
        from .robinbc import assemble_robin_bc, assemble_robin_bc_excited

        mesh = field.mesh
        w0 = 2*np.pi*frequency
        k0 = w0/C0

        er = er - 1j*sig/(w0*EPS0)*np.repeat(np.eye(3)[:, :, np.newaxis], er.shape[2], axis=2)
        
        f_dependent_properties = np.any((sig > 0) & (sig < self.conductivity_limit))
        
        if cache_matrices and not f_dependent_properties and self.cached_matrices is not None:
            logger.debug('    Retreiving cached matricies')
            E, B = self.cached_matrices
        else:
            logger.debug('    Assembling matrices')
            E, B = tet_mass_stiffness_matrices(field, er, ur)
            self.cached_matrices = (E, B)

        K: csr_matrix = (E - B*(k0**2)).tocsr()

        NF = E.shape[0]

        pecs: list[PEC] = [bc for bc in bcs if isinstance(bc,PEC)]
        robin_bcs: list[RectangularWaveguide] = [bc for bc in bcs if isinstance(bc,RobinBC)]
        ports: list[PortBC] = [bc for bc in bcs if isinstance(bc, PortBC)]
        periodic: list[Periodic] = [bc for bc in bcs if isinstance(bc, Periodic)]

        b = np.zeros((E.shape[0],)).astype(np.complex128)
        port_vectors = {port.port_number: np.zeros((E.shape[0],)).astype(np.complex128) for port in ports}
        # Process all PEC Boundary Conditions
        pec_ids = []
        
        logger.debug('    Implementing PEC Boundary Conditions.')
        
        # Conductivity above al imit, consider it all PEC
        for itet in range(field.n_tets):
            if sig[itet] > self.conductivity_limit:
                pec_ids.extend(field.tet_to_field[:,itet])
        
        # PEC Boundary conditions
        for pec in pecs:
            if len(pec.tags)==0:
                continue
            face_tags = pec.tags
            tri_ids = mesh.get_triangles(face_tags)
            edge_ids = list(mesh.tri_to_edge[:,tri_ids].flatten())

            for ii in edge_ids:
                eids = field.edge_to_field[:, ii]
                pec_ids.extend(list(eids))

            for ii in tri_ids:
                tids = field.tri_to_field[:, ii]
                pec_ids.extend(list(tids))

        # Robin BCs
        if len(robin_bcs) > 0:
            logger.debug('    Implementing Robin Boundary Conditions.')
        
            gauss_points = gaus_quad_tri(4)
            
            Bempty = field.empty_tri_matrix()

            for bc in robin_bcs:

                for tag in bc.tags:
                    face_tags = [tag,]

                    tri_ids = mesh.get_triangles(face_tags)
                    nodes = mesh.get_nodes(face_tags)
                    edge_ids = list(mesh.tri_to_edge[:,tri_ids].flatten())

                    gamma = bc.get_gamma(k0)

                    def Ufunc(x,y): 
                        return bc.get_Uinc(x,y,k0)
                    
                    ibasis = bc.get_inv_basis()
                    if ibasis is None:
                        basis = plane_basis_from_points(mesh.nodes[:,nodes]) + 1e-16
                        ibasis = np.linalg.pinv(basis)
                    if bc._include_force:

                        Bempty, b_p = assemble_robin_bc_excited(field, Bempty, tri_ids, Ufunc, gamma, ibasis, bc.cs.origin, gauss_points)
                        
                        port_vectors[bc.port_number] += b_p
                    
                    else:
                        Bempty = assemble_robin_bc(field, Bempty, tri_ids, gamma)
            B_p = field.generate_csr(Bempty)
            K = K + B_p
        
        if len(periodic) > 0:
            logger.debug('    Implementing Periodic Boundary Conditions.')

        
        # Periodic BCs
        Pmats = []
        remove = set()
        has_periodic = False

        for bc in periodic:
            Pmat = eye(NF, format='lil', dtype=np.complex128)
            has_periodic = True
            tri_ids_1 = mesh.get_triangles(bc.face1.tags)
            tri_ids_2 = mesh.get_triangles(bc.face2.tags)
            dv = np.array(bc.dv)
            trimapdict = defaultdict(lambda: [None, None])
            edgemapdict = defaultdict(lambda: [None, None])
            mult = int(10**(-np.round(np.log10(min(mesh.edge_lengths.flatten())))+3))
            
            edge_ids_1 = set()
            edge_ids_2 = set()
            
            phi = bc.phi(k0)
            for i1, i2 in zip(tri_ids_1, tri_ids_2):
                trimapdict[gen_key(mesh.tri_centers[:,i1], mult)][0] = i1
                trimapdict[gen_key(mesh.tri_centers[:,i2]-dv, mult)][1] = i2
                
                ie1, ie2, ie3 = mesh.tri_to_edge[:,i1]
                edge_ids_1.update({ie1, ie2, ie3})
                ie1, ie2, ie3 = mesh.tri_to_edge[:,i2]
                edge_ids_2.update({ie1, ie2, ie3})
            
            for i1, i2 in zip(list(edge_ids_1), list(edge_ids_2)):
                edgemapdict[gen_key(mesh.edge_centers[:,i1], mult)][0] = i1
                edgemapdict[gen_key(mesh.edge_centers[:,i2]-dv, mult)][1] = i2

            
            for t1, t2 in trimapdict.values():
                f11, f21 = field.tri_to_field[[3,7],t1]
                f12, f22 = field.tri_to_field[[3,7],t2]
                Pmat[f12,f12] = 0
                Pmat[f22,f22] = 0
                if Pmat[f12,f11] == 0:
                    Pmat[f12,f11] += phi
                else:
                    Pmat[f12,f11] *= phi
                
                if Pmat[f22,f21] == 0:
                    Pmat[f22,f21] += phi
                else:
                    Pmat[f22,f21] *= phi
                remove.add(f12)
                remove.add(f22)
                
            for e1, e2 in edgemapdict.values():
                f11, f21 = field.edge_to_field[:,e1]
                f12, f22 = field.edge_to_field[:,e2]
                Pmat[f12,f12] = 0
                Pmat[f22,f22] = 0
                if Pmat[f12,f11] == 0:
                    Pmat[f12,f11] += phi
                else:
                    Pmat[f12,f11] *= phi
                
                if Pmat[f22,f21] == 0:
                    Pmat[f22,f21] += phi
                else:
                    Pmat[f22,f21] *= phi
                remove.add(f12)
                remove.add(f22)
                
            Pmats.append(Pmat)
        
        if Pmats:
            Pmat = Pmats[0]
            for P2 in Pmats[1:]:
                Pmat = Pmat @ P2
            Pmat = Pmat.tocsr()
            remove = np.array(sorted(list(remove)))
            all_indices = np.arange(NF)
            keep_indices = np.setdiff1d(all_indices, remove)
            Pmat = Pmat[:,keep_indices]
        else:
            Pmat = None
        
        pec_ids = set(pec_ids)
        solve_ids = np.array([i for i in range(E.shape[0]) if i not in pec_ids])
        
        if has_periodic:
            mask = np.zeros((NF,))
            mask[solve_ids] = 1
            mask = mask[keep_indices]
            solve_ids = np.argwhere(mask==1).flatten()

        logger.debug(f'Number of tets: {mesh.n_tets}')
        logger.debug(f'Number of DoF: {K.shape[0]}')
        simjob = SimJob(K, b, k0*299792458/(2*np.pi), True)
        
        simjob.port_vectors = port_vectors
        simjob.solve_ids = solve_ids

        if has_periodic:
            simjob.P = Pmat
            simjob.Pd = Pmat.getH()
            simjob.has_periodic = has_periodic

        return simjob
    
    def assemble_eig_matrix(self, field: Nedelec2, 
                        er: np.ndarray, 
                        ur: np.ndarray, 
                        sig: np.ndarray,
                        bcs: list[BoundaryCondition],
                        frequency: float) -> SimJob:
        
        from .curlcurl import tet_mass_stiffness_matrices
        from .robinbc import assemble_robin_bc

        mesh = field.mesh
        w0 = 2*np.pi*frequency
        k0 = w0/C0

        er = er - 1j*sig/(w0*EPS0)*np.repeat(np.eye(3)[:, :, np.newaxis], er.shape[2], axis=2)
        
        logger.debug('    Assembling matrices')
        E, B = tet_mass_stiffness_matrices(field, er, ur)
        self.cached_matrices = (E, B)

        NF = E.shape[0]

        pecs: list[PEC] = [bc for bc in bcs if isinstance(bc,PEC)]
        robin_bcs: list[RectangularWaveguide] = [bc for bc in bcs if isinstance(bc,RobinBC)]
        periodic: list[Periodic] = [bc for bc in bcs if isinstance(bc, Periodic)]

        # Process all PEC Boundary Conditions
        pec_ids = []
        
        logger.debug('    Implementing PEC Boundary Conditions.')
        
        # Conductivity above a limit, consider it all PEC
        for itet in range(field.n_tets):
            if sig[itet] > self.conductivity_limit:
                pec_ids.extend(field.tet_to_field[:,itet])
        
        # PEC Boundary conditions
        for pec in pecs:
            if len(pec.tags)==0:
                continue
            face_tags = pec.tags
            tri_ids = mesh.get_triangles(face_tags)
            edge_ids = list(mesh.tri_to_edge[:,tri_ids].flatten())

            for ii in edge_ids:
                eids = field.edge_to_field[:, ii]
                pec_ids.extend(list(eids))

            for ii in tri_ids:
                tids = field.tri_to_field[:, ii]
                pec_ids.extend(list(tids))

        # Robin BCs
        if len(robin_bcs) > 0:
            logger.debug('    Implementing Robin Boundary Conditions.')
        
        for bc in robin_bcs:
            for tag in bc.tags:
                face_tags = [tag,]#bc.tags

                tri_ids = mesh.get_triangles(face_tags)
                nodes = mesh.get_nodes(face_tags)
                edge_ids = list(mesh.tri_to_edge[:,tri_ids].flatten())

                gamma = bc.get_gamma(k0)
                
                ibasis = bc.get_inv_basis()
                if ibasis is None:
                    basis = plane_basis_from_points(mesh.nodes[:,nodes]) + 1e-16
                    ibasis = np.linalg.pinv(basis)
                B_p = assemble_robin_bc(field, tri_ids, gamma)
                if bc._include_stiff:
                    B = B + B_p
        
        if len(periodic) > 0:
            logger.debug('    Implementing Periodic Boundary Conditions.')

        # Periodic BCs
        Pmats = []
        remove = set()
        has_periodic = False

        for bc in periodic:
            Pmat = eye(NF, format='lil', dtype=np.complex128)
            has_periodic = True
            tri_ids_1 = mesh.get_triangles(bc.face1.tags)
            tri_ids_2 = mesh.get_triangles(bc.face2.tags)
            dv = np.array(bc.dv)
            trimapdict = defaultdict(lambda: [None, None])
            edgemapdict = defaultdict(lambda: [None, None])
            mult = int(10**(-np.round(np.ceil(min(mesh.edge_lengths.flatten())))+2))
            edge_ids_1 = set()
            edge_ids_2 = set()
            
            phi = bc.phi(k0)
            for i1, i2 in zip(tri_ids_1, tri_ids_2):
                trimapdict[gen_key(mesh.tri_centers[:,i1], mult)][0] = i1
                trimapdict[gen_key(mesh.tri_centers[:,i2]-dv, mult)][1] = i2
                
                ie1, ie2, ie3 = mesh.tri_to_edge[:,i1]
                edge_ids_1.update({ie1, ie2, ie3})
                ie1, ie2, ie3 = mesh.tri_to_edge[:,i2]
                edge_ids_2.update({ie1, ie2, ie3})
            
            for i1, i2 in zip(list(edge_ids_1), list(edge_ids_2)):
                edgemapdict[gen_key(mesh.edge_centers[:,i1], mult)][0] = i1
                edgemapdict[gen_key(mesh.edge_centers[:,i2]-dv, mult)][1] = i2

            
            for t1, t2 in trimapdict.values():
                f11, f21 = field.tri_to_field[[3,7],t1]
                f12, f22 = field.tri_to_field[[3,7],t2]
                Pmat[f12,f12] = 0
                Pmat[f22,f22] = 0
                if Pmat[f12,f11] == 0:
                    Pmat[f12,f11] += phi
                else:
                    Pmat[f12,f11] *= phi
                
                if Pmat[f22,f21] == 0:
                    Pmat[f22,f21] += phi
                else:
                    Pmat[f22,f21] *= phi
                remove.add(f12)
                remove.add(f22)
                
            for e1, e2 in edgemapdict.values():
                f11, f21 = field.edge_to_field[:,e1]
                f12, f22 = field.edge_to_field[:,e2]
                Pmat[f12,f12] = 0
                Pmat[f22,f22] = 0
                if Pmat[f12,f11] == 0:
                    Pmat[f12,f11] += phi
                else:
                    Pmat[f12,f11] *= phi
                
                if Pmat[f22,f21] == 0:
                    Pmat[f22,f21] += phi
                else:
                    Pmat[f22,f21] *= phi
                remove.add(f12)
                remove.add(f22)
                
            Pmats.append(Pmat)
        
        if Pmats:
            Pmat = Pmats[0]
            for P2 in Pmats[1:]:
                Pmat = Pmat @ P2
            Pmat = Pmat.tocsr()
            remove = np.array(sorted(list(remove)))
            all_indices = np.arange(NF)
            keep_indices = np.setdiff1d(all_indices, remove)
            Pmat = Pmat[:,keep_indices]
        else:
            Pmat = None
        
        pec_ids = set(pec_ids)
        solve_ids = np.array([i for i in range(E.shape[0]) if i not in pec_ids])
        
        if has_periodic:
            mask = np.zeros((NF,))
            mask[solve_ids] = 1
            mask = mask[keep_indices]
            solve_ids = np.argwhere(mask==1).flatten()

        logger.debug(f'Number of tets: {mesh.n_tets}')
        logger.debug(f'Number of DoF: {E.shape[0]}')
        simjob = SimJob(E, None, frequency, False, B=B)
        
        simjob.solve_ids = solve_ids

        if has_periodic:
            simjob.P = Pmat
            simjob.Pd = Pmat.getH()
            simjob.has_periodic = has_periodic

        return simjob


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from typing import TypeVar
from ..geometry import GeoSurface, GeoVolume
from ..cs import CoordinateSystem, GCS
import gmsh
import numpy as np

T = TypeVar('T', GeoSurface, GeoVolume)

def _gen_mapping(obj_in, obj_out) -> dict:
    tag_mapping: dict[int, dict] = {0: dict(),
                                        1: dict(),
                                        2: dict(),
                                        3: dict()}
    for domain, mapping in zip(obj_in, obj_out):
        tag_mapping[domain[0]][domain[1]] = [o[1] for o in mapping]
    return tag_mapping

def add(main: T, tool: T, 
             remove_object: bool = True,
             remove_tool: bool = True) -> T:
    ''' Adds two GMSH objects together, returning a new object that is the union of the two.
    
    Parameters
    ----------
    main : GeoSurface | GeoVolume
    tool : GeoSurface | GeoVolume
    remove_object : bool, optional
        If True, the main object will be removed from the model after the operation. Default is True.
    remove_tool : bool, optional
        If True, the tool object will be removed from the model after the operation. Default is True.
    
    Returns
    -------
    GeoSurface | GeoVolume
        A new object that is the union of the main and tool objects.
    '''
    out_dim_tags, out_dim_tags_map = gmsh.model.occ.fuse(main.dimtags, tool.dimtags, removeObject=remove_object, removeTool=remove_tool)
    if out_dim_tags[0][0] == 3:
        output = GeoVolume([dt[1] for dt in out_dim_tags])._take_tools(tool,main)
    elif out_dim_tags[0][0] == 2:
        output = GeoSurface([dt[1] for dt in out_dim_tags])._take_tools(tool,main)
    if remove_object:
        main._exists = False
    if remove_tool:
        tool._exists = False
    return output

def remove(main: T, tool: T, 
             remove_object: bool = True,
             remove_tool: bool = True) -> T:
    ''' Subtractes a tool object GMSH from the main object, returning a new object that is the difference of the two.
    
    Parameters
    ----------
    main : GeoSurface | GeoVolume
    tool : GeoSurface | GeoVolume
    remove_object : bool, optional
        If True, the main object will be removed from the model after the operation. Default is True.
    remove_tool : bool, optional
        If True, the tool object will be removed from the model after the operation. Default is True.
    
    Returns
    -------
    GeoSurface | GeoVolume
        A new object that is the difference of the main and tool objects.
    '''
    out_dim_tags, out_dim_tags_map = gmsh.model.occ.cut(main.dimtags, tool.dimtags, removeObject=remove_object, removeTool=remove_tool)
    if out_dim_tags[0][0] == 3:
        output = GeoVolume([dt[1] for dt in out_dim_tags])._take_tools(tool,main)
    elif out_dim_tags[0][0] == 2:
        output = GeoSurface([dt[1] for dt in out_dim_tags])._take_tools(tool,main)
    if remove_object:
        main._exists = False
    if remove_tool:
        tool._exists = False
    return output

subtract = remove

def intersect(main: T, tool: T, 
             remove_object: bool = True,
             remove_tool: bool = True) -> T:
    ''' Intersection of a tool object GMSH with the main object, returning a new object that is the intersection of the two.
    
    Parameters
    ----------
    main : GeoSurface | GeoVolume
    tool : GeoSurface | GeoVolume
    remove_object : bool, optional
        If True, the main object will be removed from the model after the operation. Default is True.
    remove_tool : bool, optional
        If True, the tool object will be removed from the model after the operation. Default is True.
    
    Returns
    -------
    GeoSurface | GeoVolume
        A new object that is the difference of the main and tool objects.
    '''
    out_dim_tags, out_dim_tags_map = gmsh.model.occ.intersect(main.dimtags, tool.dimtags, removeObject=remove_object, removeTool=remove_tool)
    if out_dim_tags[0][0] == 3:
        output = GeoVolume([dt[1] for dt in out_dim_tags])._take_tools(tool,main)
    elif out_dim_tags[0][0] == 2:
        output = GeoSurface([dt[1] for dt in out_dim_tags])._take_tools(tool,main)
    if remove_object:
        main._exists = False
    if remove_tool:
        tool._exists = False
    return output

def embed(main: GeoVolume, other: GeoSurface) -> None:
    ''' Embeds a surface into a volume in the GMSH model.
    Parameters
    ----------
    main : GeoVolume
        The volume into which the surface will be embedded.
    other : GeoSurface
        The surface to be embedded into the volume.
    
    Returns
    -------
    None
    '''
    gmsh.model.geo.synchronize()
    gmsh.model.mesh.embed(other.dim, [other.tag,], main.dim, main.tags)

def rotate(main: GeoVolume, 
           c0: tuple[float, float, float],
           ax: tuple[float, float, float],
           angle: float,
           degree=True) -> GeoVolume:
    """Rotates a GeoVolume object around an axist defined at a coordinate.

    Args:
        main (GeoVolume): The object to rotate
        c0 (tuple[float, float, float]): The point of origin for the rotation axis
        ax (tuple[float, float, float]): A vector defining the rotation axis
        angle (float): The angle in degrees (if degree is True)
        degree (bool, optional): Whether to interpret the angle in degrees. Defaults to True.

    Returns:
        GeoVolume: The rotated GeoVolume object.
    """
    if degree:
        angle = angle * np.pi/180
    gmsh.model.occ.rotate(main.dimtags, *c0, *ax, -angle)

    # Rotate the facepointers
    for fp in main._all_pointers:
        fp.rotate(c0, ax, angle)
    return main

def translate(main: GeoVolume,
              dx: float = 0,
              dy: float = 0,
              dz: float = 0) -> GeoVolume:
    """Translates the GeoVolume object along a given displacement

    Args:
        main (GeoVolume): The object to translate
        dx (float, optional): The X-displacement in meters. Defaults to 0.
        dy (float, optional): The Y-displacement in meters. Defaults to 0.
        dz (float, optional): The Z-displacement in meters. Defaults to 0.

    Returns:
        GeoVolume: The translated object
    """
    gmsh.model.occ.translate(main.dimtags, dx, dy, dz)
    
     # Rotate the facepointers
    for fp in main._all_pointers:
        fp.translate(dx, dy, dz)

    return main

def mirror(main: GeoVolume,
           origin: tuple[float, float, float] = (0.0, 0.0, 0.0),
           direction: tuple[float, float, float] = (1.0, 0.0, 0.0),
           make_copy: bool = True) -> GeoVolume:
    """Mirrors a GeoVolume object along a miror plane defined by a direction originating at a point

    Args:
        main (GeoVolume): The object to mirror
        origin (tuple[float, float, float], optional): The point of origin in meters. Defaults to (0.0, 0.0, 0.0).
        direction (tuple[float, float, float], optional): The normal axis defining the plane of reflection. Defaults to (1.0, 0.0, 0.0).

    Returns:
        GeoVolume: The mirrored GeoVolume object
    """
    a, b, c = direction
    x0, y0, z0 = origin
    d = -(a*x0 + b*y0 + c*z0)
    if (a==0) and (b==0) and (c==0):
        return main
    mirror_obj = main
    if make_copy:
        new_obj = main.make_copy()
        gmsh.model.occ.mirror(new_obj.dimtags, a,b,c,d)
        mirror_obj = new_obj
    else:
        gmsh.model.occ.mirror(main.dimtags, a,b,c,d)

    for fp in mirror_obj._all_pointers:
        fp.mirror(origin, direction)
    return mirror_obj

def change_coordinate_system(main: GeoVolume,
                             new_cs: CoordinateSystem = GCS,
                             old_cs: CoordinateSystem = GCS):
    """Moves the GeoVolume object from a current coordinate system to a new one.

    The old and new coordinate system by default are the global coordinate system.
    Thus only one needs to be provided to transform to and from these coordinate systems.

    Args:
        main (GeoVolume): The object to transform
        new_cs (CoordinateSystem): The new coordinate system. Defaults to GCS
        old_cs (CoordinateSystem, optional): The old coordinate system. Defaults to GCS.

    Returns:
        _type_: _description_
    """
    if new_cs._is_global and old_cs._is_global:
        return main
    
    M1 = old_cs.affine_to_global()
    M2 = new_cs.affine_from_global()
    # Transform to the global coordinate system.
    if not old_cs._is_global:
        gmsh.model.occ.affine_transform(main.dimtags, M1.flatten()[:12])
    # Transform to a new coordinate system.
    if not new_cs._is_global:
        gmsh.model.occ.affineTransform(main.dimtags, M2.flatten()[:12])

    for fp in main._all_pointers:
        fp.affine_transform(M1)
        fp.affine_transform(M2)
    return main


import sys
import argparse
from _emerge.projects.generate_project import generate_project

def main():
    parser = argparse.ArgumentParser(description="Emerge Project Generator CLI")
    subparsers = parser.add_subparsers(dest="command")

    # Subcommand: new
    new_parser = subparsers.add_parser("new", help="Create a new project")
    new_parser.add_argument("projectname", type=str, help="Name of the project directory")
    new_parser.add_argument("filename", type=str, help="Base name for files")

    args = parser.parse_args()

    if args.command == "new":
        generate_project(args.projectname, args.filename)
    else:
        parser.print_help()



# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

import numpy as np
import os
from scipy.sparse import csr_matrix, save_npz, load_npz
from ...solver import SolveReport

class SimJob:

    def __init__(self, 
                 A: csr_matrix,
                 b: np.ndarray,
                 freq: float,
                 cache_factorization: bool,
                 B: csr_matrix = None
                 ):

        self.A: csr_matrix = A
        self.B: csr_matrix = B
        self.b: np.ndarray = b
        self.P: csr_matrix = None
        self.Pd: csr_matrix = None
        self.has_periodic: bool = False

        self.freq: float = freq
        self.k0: float = 2*np.pi*freq/299792458
        self.cache_factorization: bool = cache_factorization
        self._fields: dict[int, np.ndarray] = dict()
        self.port_vectors: dict = None
        self.solve_ids = None

        self.store_limit = None
        self.relative_path = None
        self._store_location = {}
        self._stored: bool = False

        self._active_port: int = None
        self.reports: list[SolveReport] = []
        self.id: int = -1
        self.store_if_needed()

    def maybe_store(self, matrix, name):
        if self.store_limit is None:
            return matrix
        
        if matrix is not None and matrix.nnz > self.store_limit:
            # Create temp directory if needed
            os.makedirs(self.relative_path, exist_ok=True)
            path = os.path.join(self.relative_path, f"csr_{str(self.freq).replace('.','_')}_{name}.npz")
            save_npz(path, matrix, compressed=False)
            self._store_location[name] = path
            self._stored = True
            return None  # Unload from memory
        return matrix
    
    def store_if_needed(self):
        self.A = self.maybe_store(self.A, 'A')
        if self.has_periodic:
            self.P = self.maybe_store(self.P, 'P')
            self.Pd = self.maybe_store(self.Pd, 'Pd')

    def load_if_needed(self, name):
        if name in self._store_location:
            return load_npz(self._store_location[name])
        return getattr(self, name)

    def iter_Ab(self):
        reuse_factorization = False

        for key, mode in self.port_vectors.items():
            # Set port as active and add the port mode to the forcing vector
            self._active_port = key

            b_active = self.b + mode
            A = self.load_if_needed('A')
            

            if self.has_periodic:
                P = self.load_if_needed('P')
                Pd = self.load_if_needed('Pd')
                b_active = Pd @ b_active
                A = Pd @ A @ P

            yield A, b_active, self.solve_ids, reuse_factorization

            reuse_factorization = True
        
        self.cleanup()

    def yield_AC(self):
        A = self.A
        B = self.B
        
        if self.has_periodic:
            P = self.P
            Pd = self.Pd
            A = Pd @ A @ P
            B = Pd @ B @ P

        return A, B, self.solve_ids
        
    def fix_solutions(self, solution: np.ndarray) -> np.ndarray:
        if self.has_periodic:
            solution = self.P @ solution
        return solution
    
    def submit_solution(self, solution: np.ndarray, report: SolveReport):
        # Solve the Ax=b problem

        if self.has_periodic:
            solution = self.P @ solution
        # From now reuse the factorization

        self._fields[self._active_port] = solution
        self.reports.append(report)
        self.routine = None
    
    def cleanup(self):
        if not self._stored:
            return
        
        if not os.path.isdir(self.relative_path):
            return

        # Remove only the files we saved
        for path in self._store_location.values():
            if os.path.isfile(path):
                os.remove(path)

        # If the directory is now empty, remove it
        if not os.listdir(self.relative_path):
            os.rmdir(self.relative_path)


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

import numpy as np

def  norm(Field: np.ndarray) -> np.ndarray:
    ''' Computes the complex norm of a field (3,N)'''
    return np.sqrt(np.abs(Field[0,:])**2 + np.abs(Field[1,:])**2 + np.abs(Field[2,:])**2)

def coax_rout(rin: float,
              eps_r: float = 1,
              Z0: float = 50) -> float:
    return rin*10**(Z0*np.sqrt(eps_r)/138)

def coax_rin(rout: float,
              eps_r: float = 1,
              Z0: float = 50) -> float:
    return rout/10**(Z0*np.sqrt(eps_r)/138)




# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from __future__ import annotations
import numpy as np
from typing import Callable, Iterable, Any
from .shapes import Box, Sphere, Cyllinder, CoaxCyllinder, Alignment
from ..geometry import GeoVolume, GeoObject, GeoSurface
from .operations import rotate, mirror, translate, add, subtract, embed
from numbers import Number
from functools import reduce
from operator import mul
from ..cs import CoordinateSystem, GCS
from .polybased import XYPolygon

def get_flat_index(indices, shape):
    flat_index = 0
    for i, idx in enumerate(indices):
        stride = reduce(mul, shape[i+1:], 1)
        flat_index += idx * stride
    return flat_index


class Series:

    def __init__(self, values: Iterable[Number]):
        self.values: list[Number] = list(values)
        self.N: int = len(self.values)

    def __len__(self) -> int:
        return self.N
    
    def __getitem__(self, index: int):
        return self.values[index]

def get_count(args) -> int:
    N = 1
    for arg in args:
        if isinstance(arg,Series):
            N = max(N, arg.N)
            return N
        elif isinstance(arg, Iterable):
            return get_count(arg)
    return 1

def get_item(value, n: int):
    if isinstance(value, Number):
        return value
    elif isinstance(value, Series):
        return value[n]
    elif isinstance(value, Iterable):
        typ = type(value)
        return typ([get_item(v,n) for v in value])
    return None

def unpack(*args):
    output = []
    N = get_count(args)
    return tuple(zip(*[[get_item(a,n) for a in args] for n in range(N)]))

ObjectTransformer = Callable[[GeoObject,], GeoObject]

class PartialFunction:

    def __init__(self, 
                 function: Callable, 
                 identity: bool = False,
                 permute: bool = False,
                 **kwargs):
        self.f: Callable = function
        self.kwargs: dict[str, float | Series] = kwargs
        self.identity: bool = identity
        self.permute: bool = permute
        self.N: int = 1
        self.kwargset: list[dict] = None

        for key, value in self.kwargs.items():
            if not isinstance(value, Series):
                continue
            self.N = max(self.N, value.N)
        
        # Include the identity operation
        if self.identity:
            self.N = self.N + 1
    
    def _compile(self) -> None:
        kwargset = []
        N = self.N

        if self.identity:
            kwargset.append(None)
            N = N - 1
        
        for n in range(N):
            kwargs = dict()
            for key,value in self.kwargs.items():
                if not isinstance(value, Series):
                    kwargs[key] = value
                else:
                    kwargs[key] = value[n]
            kwargset.append(kwargs)
        self.kwargset = kwargset

    def __call__(self, objects: list[GeoObject], index: int) -> list[GeoObject]:
        
        kwargs = self.kwargset[index]
        if kwargs is None:
            return objects
        
        objects_out = []
        for obj in objects:
            objects_out.append(self.f(obj, **kwargs))
        return objects_out

class NDimContainer:

    def __init__(self):
        self.items: list= []
        self.dimensions: list[tuple[int,int]] = []
        self.expanded_dims: list[tuple[int,int,int]] = []
        self.map: np.ndarray = []
        self.Ncopies: int = 1

    @property
    def Ntot(self) -> int:
        """The Total number of different objects this NDimContainer acts on.
        """
        n = self.Ncopies
        for i,N in self.dimensions:
            n = n*N
        
        return n

    @property
    def Nop(self) -> int:
        """The number of operations corresponding to this NDimContainer

        Returns:
            int: The number of operations this container can support.
        """
        n = 1
        for i,N in self.dimensions:
            n = n*N
        
        return n
    
    @property
    def dimtup(self) -> tuple[int]:
        """A dimension tuple containing the dimensions of the NDimContainer

        Returns:
            tuple[int]: A tuple of dimension sizes to be used in np.zeros(shape)
        """
        dimtup =  tuple([dim[1] for dim in self.dimensions])
        if self.Ncopies:
            dimtup = dimtup + (self.Ncopies,)
        return dimtup
    
    def _init(self) -> None:
        """Initialization function to set the index map and expanded dimension list.
        """
        i = 0
        for i,(ncopies,N) in enumerate(self.dimensions):
            for n in range(ncopies):
                self.expanded_dims.append((i,n,N))
        
        if self.Ncopies > 1:
            self.expanded_dims.append((i+1, (1, self.Ncopies)))

        self.map = np.arange(self.Ntot).reshape(self.dimtup)

    def add_dim(self, N: int, same_axis: bool = False) -> None:
        """Aads a new dimension to iterate over.

        Args:
            N (int): The size of the dimension (number of differnt items contained)
            same_axis (bool, optional): Defines that the next iteration dimension is actually parallel
            to the last existing dimension. In this case N must be equal to the size of the last defined dimension. Defaults to False.

        Raises:
            ValueError: An error if the provided dimension is not correct.
        """
        if same_axis:
            if N != self.dimensions[-1][1]:
                raise ValueError('Trying to add a dimension with the same size as previous but the sizes are different.' + 
                                 f'The provided size is {N} but the last dimensions is {self.dimensions[-1][1]}')
            self.dimensions[-1] = (self.dimensions[-1][0]+1, N)
        else:
            self.dimensions.append((1,N))

    def set_copies(self, N: int) -> None:
        """Define how many original copies of the new object will be created.

        Args:
            N (int): The number of copies.
        """
        self.Ncopies = N

    def get(self, 
            dim_ex_dim: int,
            number: int) -> list:
        dimindex, niter, Nvariations = self.expanded_dims[dim_ex_dim]
        slclist = [slice(None) for _ in self.dimensions]
        slclist[dimindex] = number
        return list(self.map[tuple(slclist)].flatten())
        
        
class Modeler:
    
    def __init__(self):
        ## Function importing
        self.rotate = rotate
        self.add = add
        self.translate = translate
        self.remove = subtract
        self.subtract = subtract
        self.embed = embed
        self.mirror = mirror

        self.pre_transforms: list[PartialFunction] = []
        self.post_transforms: list[PartialFunction] = []
        self.last_object: GeoObject = None
        self.ndimcont: NDimContainer = NDimContainer()
        self._and: bool = False

    @property
    def AND(self) -> Modeler:
        """This method chain property ensures that the following transformation will be merged
        with the previous one. For example, a rotation plus a translation with AND in between
        both containing three operations will merge their operations so that each single value
        is paired up with the next. 

        For example, if you wish to create a three boxes that are rotated in increments of 90
        degrees and traslated by 1, 2 and 3 meters respectively (90deg, 1m) (180deg, 2m) and (270deg, 3m)
        you would do:

        >>> m.rotated(...,..., m.Series(90,180,270)).AND.translated(m.Series(1,2,3),...)

        If you would not use AND, the operations are permuted resulting in 9 boxes.
        """
        self._and = True
        return self

    def series(self, *values: Number) -> Series:
        """Returns a Series object to represent a series of values instead of a single value.
        
        A series of values can be used at any argument that accepts Series objects. If such a
        variable is provided as a series, the modeler will execute multiple iterations of the 
        object or transformation for each value in the series.

        Returns:
            Series: The generated Series object.

        Example:
        >>> modeler.box(modeler.Series(1,2,3), 2, 3)

        """
        return Series(values)
    
    def nvars(self) -> int:
        return self.ndimcont.Nop
    
    def _merge_object(self, objs: list[GeoObject]) -> GeoObject:
        """Merges a list of GeoObjects into a single GeoObject.

        Args:
            objs (list[GeoObject]): A list of GMSH objects

        Returns:
            GeoObject: A single GeoObject
        """
        tags = []
        for obj in objs:
            tags.extend(obj.tags)
        gmshobj = GeoObject.from_dimtags(objs[0].dim, tags)
        for obj in objs:
            gmshobj._take_tools(obj)
        return gmshobj
    
    def _clean(self) -> None:
        self.pre_transforms = []
        self._and = False
        self._combine = False

    def _add_dim(self, N: int) -> None:
        self.ndimcont.add_dim(N, self._and)
        self._and = False

    def _add_function(self, function: PartialFunction):
        self._add_dim(function.N)
        self.pre_transforms.append(function)

    def _apply_presteps(self, objects: list[GeoObject]) -> list[GeoObject]:
        self.ndimcont._init()
        for func in self.pre_transforms:
            func._compile()
        if not self.pre_transforms:
            return objects
        
        for i, function in enumerate(self.pre_transforms):
            for n in range(function.N):
                ids = self.ndimcont.get(i,n)
                sel = [objects[i] for i in ids]
                objects_out = function(sel, n)
                for j, obj in zip(ids, objects_out):
                    objects[j] = obj

        self._clean()
        return objects
    
    def rotated(self,
                c0: tuple[float, float, float],
                ax: tuple[float, float, float],
                angle: float | Series,
                ) -> Modeler:
        """Adds a rotation task to the geometry builder.

        Args:
            c0 (tuple[float, float, float]): The origin at which the rotation axis is placed.
            ax (tuple[float, float, float]): The Rotation axis to be used
            angle (float | Series): The angles or series of angles provided in degrees.

        Returns:
            Modeler: The same modeler instance
        """
        function = PartialFunction(rotate, c0=c0, ax=ax, angle=angle)
        self._add_function(function)
        return self
    
    def translated(self,
                   dx: float | Series,
                   dy: float | Series,
                   dz: float | Series) -> Modeler:
        """Adds a translation task to the geometry builder pipeline.
        The dx, dy, dz may either be a single displacement coordinate or a series of them in the 
        shape of a series object.

        Args:
            dx (float | Series): The x-displacement
            dy (float | Series): The y-displacement
            dz (float | Series): The z-displacement

        Returns:
            Modeler: The Modeler object
        """
        function = PartialFunction(translate, dx=dx, dy=dy, dz=dz)
        self._add_function(function)
        return self
    
    def mirrored(self,
               origin: tuple[float, float, float],
               direction: tuple[float, float, float],
               keep_original: bool = True) -> Modeler:
        """Adds a mirror transformation to the geometry builder pipeline.

        Args:
            origin (tuple[float, float, float]): The origin of the mirror axis plane
            direction (tuple[float, float, float]): The mirror direction which is normal to the plane of reflection.
            keep_original (bool, optional): Whether to keep the original object. Defaults to True.

        Returns:
            Modeler: The Modeler object
        """
        function = PartialFunction(mirror, 
                                                   identity=keep_original, 
                                                   origin=origin, 
                                                   direction=direction)
        self._add_function(function)
        return self
    
    def cyllinder(self,
                  radius: float | Series,
                  height: float | Series,
                  position: tuple[float | Series, float | Series, float | Series],
                  cs: CoordinateSystem = GCS,
                  merge: bool = False,
                  NPoly: int = False):
        
        N_objects = self.nvars()
        Rs, Hs, Ps = unpack(radius, height, position)
        N = len(Rs)
        cyls = []
        
        for _ in range(N_objects):
            for r,h,p in zip(Rs, Hs, Ps):
                cs2 = cs.displace(p[0], p[1], p[2])
                if NPoly:
                    cyl = XYPolygon.circle(r, Nsections=NPoly).extrude(h, cs2)
                else:
                    cyl  = Cyllinder(r,h, cs2)
                cyls.append(cyl)

        self.ndimcont.set_copies(N)
        self._apply_presteps(cyls)

        if merge:
            cyls = self._merge_object(cyls)
        
        return cyls
    
    def box(self, width: float | Series,
            depth: float | Series,
            height: float | Series,
            position: tuple[float | Series, float | Series, float | Series],
            alignment: Alignment = Alignment.CORNER,
            merge: bool = False) -> list[GeoVolume] | GeoVolume:
        """Create a box object which will be transformed by the transformation pipeline.

        Args:
            width (float): The box's width (X direction)
            depth (float): The box's depth (Y direction)
            height (float): The box's height (Z direction)
            position (tuple[float, float, float]): The position of the box object.
            alignment (Alignment, optional): Where to alight the box. Defaults to Alignment.CORNER.
            merge (bool): Whether to merge the final result into a single GeoVolume object.

        Returns:
            list[GeoVolume]: A list of GeoVolume objects for each box.
        """
        N_objects = self.nvars()
        Ws, Ds, Hs, Ps = unpack(width, depth, height, position)
        N = len(Ws)
        boxes = []
        
        for _ in range(N_objects):
            for w, d, h, p in zip(Ws, Ds, Hs, Ps):
                box  = Box(w,d,h, position=p, alignment=alignment)
                boxes.append(box)
        self.ndimcont.set_copies(N)
        self._apply_presteps(boxes)

        if merge:
            boxes = self._merge_object(boxes)
        
        return boxes


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from ..geometry import GeoVolume
from ..cs import CoordinateSystem
from ..selection import FaceSelection

import gmsh

class Horn(GeoVolume):

    def __init__(self,
                 start_aperture: tuple[float, float],
                 end_aperture: tuple[float, float],
                 height: float,
                 cs: CoordinateSystem):
        """Generate a Horn geometry. The horn is defined by a start and ending rectangle.

        Start aperture is always defined at z=0 and the end at z=height.
        The horn can be reoriented by choosing a different coordinate system.

        Args:
            start_aperture (tuple[float, float]): The width/height of the start aperture
            end_aperture (tuple[float, float]): The width/height of the end aperture
            height (float): The height of the horn
            cs (CoordinateSystem): The coordinate systme to place the horn in.
        """
        super().__init__([])
        p0 = cs.origin
        p1 = p0 + cs.zax.np * height

        wax = cs.xax.np
        hax = cs.yax.np
        dax = cs.zax.np

        w1, h1 = start_aperture
        w2, h2 = end_aperture

        p11 = p0 + wax *w1/2 + hax * h1/2
        p12 = p0 + wax *w1/2 - hax * h1/2
        p13 = p0 - wax *w1/2 - hax * h1/2
        p14 = p0 - wax *w1/2 + hax * h1/2

        p21 = p1 + wax *w2/2 + hax * h2/2
        p22 = p1 + wax *w2/2 - hax * h2/2
        p23 = p1 - wax *w2/2 - hax * h2/2
        p24 = p1 - wax *w2/2 + hax * h2/2

        pt11 = gmsh.model.occ.addPoint(*p11)
        pt12 = gmsh.model.occ.addPoint(*p12)
        pt13 = gmsh.model.occ.addPoint(*p13)
        pt14 = gmsh.model.occ.addPoint(*p14)
        pt21 = gmsh.model.occ.addPoint(*p21)
        pt22 = gmsh.model.occ.addPoint(*p22)
        pt23 = gmsh.model.occ.addPoint(*p23)
        pt24 = gmsh.model.occ.addPoint(*p24)

        l1r = gmsh.model.occ.addLine(pt11, pt12)
        l1b = gmsh.model.occ.addLine(pt12, pt13)
        l1l = gmsh.model.occ.addLine(pt13, pt14)
        l1t = gmsh.model.occ.addLine(pt14, pt11)
        
        l2r = gmsh.model.occ.addLine(pt21, pt22)
        l2b = gmsh.model.occ.addLine(pt22, pt23)
        l2l = gmsh.model.occ.addLine(pt23, pt24)
        l2t = gmsh.model.occ.addLine(pt24, pt21)

        dtr = gmsh.model.occ.addLine(pt11, pt21)
        dbr = gmsh.model.occ.addLine(pt12, pt22)
        dbl = gmsh.model.occ.addLine(pt13, pt23)
        dtl = gmsh.model.occ.addLine(pt14, pt24)

        wbot = gmsh.model.occ.addWire([l1r, l1b, l1l, l1t])
        wtop = gmsh.model.occ.addWire([l2r, l2b, l2l, l2t])
        wright = gmsh.model.occ.addWire([l1r, dbr, l2r, dtr])
        wleft = gmsh.model.occ.addWire([l1l, dbl, l2l, dtl])
        wfront = gmsh.model.occ.addWire([l1b, dbl, l2b, dbr])
        wback = gmsh.model.occ.addWire([l1t, dtr, l2t, dtl])

        s1 = gmsh.model.occ.addSurfaceFilling(wbot)
        s2 = gmsh.model.occ.addSurfaceFilling(wtop)
        s3 = gmsh.model.occ.addSurfaceFilling(wright)
        s4 = gmsh.model.occ.addSurfaceFilling(wleft)
        s5 = gmsh.model.occ.addSurfaceFilling(wfront)
        s6 = gmsh.model.occ.addSurfaceFilling(wback)

        sv = gmsh.model.occ.addSurfaceLoop([s1, s2, s3, s4, s5, s6])

        self.tags: list[int] = [gmsh.model.occ.addVolume([sv,]),]
        
        pc = p0 + dax * height/2
        self._add_face_pointer('front', pc - height/2*dax, -dax)
        self._add_face_pointer('back', pc + height/2*dax, dax)


from __future__ import annotations
import matplotlib.pyplot as plt
import numpy as np
from typing import Any
from itertools import cycle

class Style:

    def __init__(self):
        self.linecolors: list = ['k']
        self.linestyles: list[str] = ['-']
        self.markers: list[str] = [None]

        self.color_cycler = cycle(self.linecolors)
        self.style_cycler = cycle(self.linestyles)
        self.marker_cycle = cycle(self.markers)

    def get_line_properties(self, color: str, linestyle: str, marker: str) -> dict[str, Any]:
        if color is None:
            color = next(self.color_cycler)
        if linestyle is None:
            linestyle = next(self.style_cycler)
        if marker is None:
            marker = next(self.marker_cycle)
        return dict(color=color, linestyle=linestyle, marker=marker)

    def dress(self, axis):
        axis.grid(True, axis='both', color='k')

class Grapher:

    def __init__(self):
        self.axes: list = []
        self.axes_grid: list[list] = []
        self.fig = None
        self.style: Style = Style()
        self.nrows: int = 0
        self.ncols: int = 0
        self.current: int = 0

        self.cur_axis = None
    
    @property
    def ax(self) -> plt.axis:
        if self.cur_axis is None:
            return self.axes[0]
        return self.cur_axis
    
    def reset(self):
        self.cur_axis = None

    def __call__(self, i: int = None, j: int = None) -> Grapher:
        if i is None:
            self.cur_axis = self.axes[0]
        return self
        if j is None and i is not None:
            self.cur_axis = self.axes[i]
        else:
            self.cur_axis = self.axes[i,j]
        return self
    
    def new(self, rows: int = 1, cols: int = 1) -> Grapher:
        self.fig, self.axes = plt.subplots(rows, cols)
        if rows==1 and cols==1:
            self.axes = [self.axes]
        self.nrows = rows
        self.ncols = cols 
        return self

    def line(self,
             xs: np.ndarray,
             ys: np.ndarray,
             color: str = None,
             linestyle: str = None,
             marker: str = None,
             dB: bool = False) -> Grapher:
        if dB:
            ys = 20*np.log10(np.abs(ys))
        self.ax.plot(xs, ys, **self.style.get_line_properties(color=color, linestyle=linestyle, marker=marker))
        self.reset()
        return self

    def show(self):
        for ax in self.axes:
            self.style.dress(ax)
        plt.show()

gr = Grapher().new()

xs = np.linspace(1e9, 2e9, 1001)
ys = np.sin(xs/1e8)

gr().line(xs, ys).show()


# """A Python based FEM solver.
# Copyright (C) 2025  name of Robert Fennis

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

# """
# import os

# os.environ["OMP_NUM_THREADS"] = "1"
# os.environ["MKL_NUM_THREADS"] = "1"
# os.environ["OPENBLAS_NUM_THREADS"] = "1"

# from loguru import logger
# from .logsettings import logger_format
# import sys

# logger.remove()
# logger.add(sys.stderr, format=logger_format)

# logger.debug('Importing modules')
# from _emerge.simmodel import Simulation3D
# from _emerge.material import Material, FR4, AIR, VACUUM, COPPER
# import bc
# from _emerge.solver import superlu_info, SolverBicgstab, SolverGMRES, SolveRoutine, ReverseCuthillMckee, Sorter, SolverPardiso, SolverUMFPACK
# from _emerge.cs import CoordinateSystem, Plane, Axis, XAX, YAX, ZAX, XYPLANE, XZPLANE, YZPLANE, YXPLANE, ZXPLANE, ZYPLANE
# from _emerge.coord import Line
# import geo
# from _emerge.selection import Selection, FaceSelection, DomainSelection, EdgeSelection
# from _emerge.mth.common_functions import norm
# logger.debug('Importing complete!')


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from __future__ import annotations
from ..mesh3d import Mesh3D
import numpy as np
from typing import Callable
from scipy.sparse import coo_matrix, csr_matrix

from ..mth.optimized import matmul

class FEMBasis:

    def __init__(self, mesh: Mesh3D):
        self.mesh: Mesh3D = mesh
        self.n_edges: int = self.mesh.n_edges
        self.n_tris: int = self.mesh.n_tris
        self.n_tets: int = self.mesh.n_tets

        self.n_tet_dofs: int = None
        self.n_tri_dofs: int = None

        self.n_field: int = 2*self.n_edges + 2*self.n_tris

        self.tet_to_field: np.ndarray = None
        
        self.edge_to_field: np.ndarray = None

        self.tri_to_field: np.ndarray = None

        self._rows: np.ndarray = None
        self._cols: np.ndarray = None
        
    
    def interpolate_Ef(self, field: np.ndarray, basis: np.ndarray = None, origin: np.ndarray = None, tetids: np.ndarray = None) -> Callable:
        '''Generates the Interpolation function as a function object for a given coordiante basis and origin.'''
        if basis is None:
            basis = np.eye(3)

        if origin is None:
            origin = np.zeros(3)
        
        ibasis = np.linalg.pinv(basis)
        def func(xs: np.ndarray, ys: np.ndarray, zs: np.ndarray) -> np.ndarray:
            xyz = np.array([xs, ys, zs]) + origin[:, np.newaxis]
            xyzg = matmul(basis, xyz)
            return matmul(ibasis, np.array(self.interpolate(field, xyzg[0,:], xyzg[1,:], xyzg[2,:], tetids)))
        return func
    
    def interpolate(self, field: np.ndarray, xs: np.ndarray, ys: np.ndarray, zs: np.ndarray, tetids: np.ndarray = None) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        pass
    
    def interpolate_curl(self, field: np.ndarray, xs: np.ndarray, ys: np.ndarray, zs: np.ndarray, constants: np.ndarray, tetids: np.ndarray = None) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Interpolates the curl of the field at the given points.
        """
        pass
    
    def empty_tet_matrix(self) -> np.ndarray:
        nnz = self.n_tets*self.n_tet_dofs**2
        matrix = np.empty((nnz,), dtype=np.complex128)
        return matrix
    
    def empty_tri_matrix(self) -> np.ndarray:
        nnz = self.n_tris*self.n_tri_dofs**2
        matrix = np.zeros((nnz,), dtype=np.complex128)
        return matrix
     
    def empty_tet_rowcol(self) -> tuple[np.ndarray, np.ndarray]:
        N = self.n_tet_dofs
        N2 = N**2
        nnz = self.n_tets*N2
        rows = np.empty(nnz, dtype=np.int64)
        cols = np.empty(nnz, dtype=np.int64)

        for itet in range(self.n_tets):
            p = itet*N2
            
            indices = self.tet_to_field[:, itet]
            for ii in range(self.n_tet_dofs):
                rows[p+N*ii:p+N*(ii+1)] = indices[ii]
                cols[p+ii:p+N2:N] = indices[ii]
        self._rows = rows
        self._cols = cols 
        return rows, cols
    
    def empty_tri_rowcol(self) -> tuple[np.ndarray, np.ndarray]:
        N = self.n_tri_dofs
        N2 = N**2
        nnz = self.n_tris*N2
        rows = np.empty(nnz, dtype=np.int64)
        cols = np.empty(nnz, dtype=np.int64)

        for itri in range(self.n_tris):
            p = itri*N2
            indices = self.tri_to_field[:, itri]
            for ii in range(N):
                rows[p+N*ii:p+N*(ii+1)] = indices[ii]
                cols[p+ii:p+N2:N] = indices[ii]

        self._rows = rows
        self._cols = cols
        return rows, cols

    def tetslice(self, itet: int) -> slice:
        N = self.n_tet_dofs**2
        return slice(itet*N,(itet+1)*N)
    
    def trislice(self, itri: int) -> slice:
        N = self.n_tri_dofs**2
        return slice(itri*N,(itri+1)*N)
    
    def generate_csr(self, data: np.ndarray) -> csr_matrix:
        ids = np.argwhere(data!=0)[:,0]
        return csr_matrix((data[ids], (self._rows[ids], self._cols[ids])), shape=(self.n_field, self.n_field))
    ### QUANTITIES

    def tet_to_edge_lengths(self, itet: int) -> np.ndarray:
        """
        Returns the edge lengths of the tetrahedron itet.
        """
        return self.mesh.edge_lengths[self.mesh.tet_to_edge[:,itet]]
    
    def tri_to_edge_lengths(self, itri: int) -> np.ndarray:
        """
        Returns the edge lengths of the triangle itri.
        """
        return self.mesh.edge_lengths[self.mesh.tri_to_edge[:,itri]]
    ####### INDEX MAPPINGS

    def local_tet_to_triid(self, itet: int) -> np.ndarray:
        raise NotImplementedError("local_tet_to_triid not implemented")
    
    def local_tet_to_edgeid(self, itet: int) -> np.ndarray:
        raise NotImplementedError("local_tet_to_edgeid not implemented")

    def interpolate_index(self, xs: np.ndarray,
                    ys: np.ndarray,
                    zs: np.ndarray,
                    tetids: np.ndarray = None) -> np.ndarray:
        raise NotImplementedError()
    
    def map_edge_to_field(self, edge_ids: np.ndarray) -> np.ndarray:
        """
        Returns the field ids for the edges.
        """
        return edge_ids

    # ##### CUTS

    @property
    def bounds(self) -> tuple[tuple[float, float], tuple[float, float],tuple[float, float]]:
        xmin = np.min(self.mesh.nodes[0,:])
        xmax = np.max(self.mesh.nodes[0,:])
        ymin = np.min(self.mesh.nodes[1,:])
        ymax = np.max(self.mesh.nodes[1,:])
        zmin = np.min(self.mesh.nodes[2,:])
        zmax = np.max(self.mesh.nodes[2,:])
        return (xmin, xmax), (ymin, ymax), (zmin, zmax)

    # @staticmethod
    # def tet_stiff_mass_submatrix(tet_vertices: np.ndarray, 
    #                              edge_lengths: np.ndarray, 
    #                              local_edge_map: np.ndarray, 
    #                              local_tri_map: np.ndarray, 
    #                              C_stiffness: float, 
    #                              C_mass: float) -> tuple[np.ndarray, np.ndarray]:
    #     pass
    
    # @staticmethod
    # def tri_stiff_mass_submatrix(tri_vertices: np.ndarray, 
    #                              edge_lengths: np.ndarray,
    #                              local_edge_map: np.ndarray,
    #                              C_stiffness: float, 
    #                              C_mass: float) -> tuple[np.ndarray, np.ndarray]:
    #     pass
    
    # @staticmethod
    # def tri_stiff_vec_matrix(lcs_vertices: np.ndarray, 
    #                          edge_lengths: np.ndarray, 
    #                          gamma: complex, 
    #                          lcs_Uinc: np.ndarray, 
    #                          DPTs: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    #     pass

    # @staticmethod
    # def tri_stiff_matrix(lcs_vertices: np.ndarray, 
    #                          edge_lengths: np.ndarray, 
    #                          gamma: complex) -> tuple[np.ndarray, np.ndarray]:
    #     pass
    
    
    # @staticmethod
    # def tri_surf_integral(lcs_vertices: np.ndarray, 
    #                       edge_lengths: np.ndarray, 
    #                       lcs_Uinc: np.ndarray, 
    #                       DPTs: np.ndarray) -> complex:
    #     pass



# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from __future__ import annotations
import numpy as np
from typing import Callable

def gauss3_composite(x: np.ndarray, y: np.ndarray) -> float:
    """
    Composite 3-point Gauss (order-2) integral
    on *equally spaced* 1-D data.

    Sections:
        (x0,x1,x2), (x2,x3,x4), (x4,x5,x6), ...

    The rule on one section [x_i, x_{i+2}] is

        ∫ f dx  ≈  (h/2) Σ_k w_k · f( x̄ + (h/2) ξ_k )

    where  h = x_{i+2}−x_i  and
          ξ = (−√3/5, 0, +√3/5),
          w = (8/9, 5/9, 8/9)  ← as requested.

    Because f is only known at the three nodes, we
    rebuild the quadratic that interpolates them and
    evaluate that polynomial at the Gauss points.
    """
    x = np.asarray(x, dtype=np.complex128)
    y = np.asarray(y, dtype=np.complex128)

    if x.ndim != 1 or y.ndim != 1 or x.size != y.size:
        raise ValueError("x and y must be 1-D arrays of equal length.")
    if x.size % 2 == 0:
        raise ValueError("Number of samples must be odd (… 0,1,2; 2,3,4; …).")

    # constant spacing
    h = x[1] - x[0]

    # Gauss–Legendre nodes and *your* weights
    xi = np.sqrt(3/5)
    nodes   = np.array([-xi, 0.0, +xi])
    weights = np.array([5/9, 8/9, 5/9])

    total = 0.0
    for i in range(0, x.size - 2, 2):
        y0, y1, y2 = y[i:i+3]
    #     array([[ 5.00000000e-01, -1.00000000e+00,  5.00000000e-01],
    #    [-5.00000000e-01,  3.71914213e-17,  5.00000000e-01],
    #    [-5.55111512e-17,  1.00000000e+00,  5.55111512e-17]])
        # coefficients of the quadratic passing through
        # (-1,y0), (0,y1), (1,y2) in local coords t
        a = y0*0.5 - y1 + 0.5*y2
        b = -y0*0.5 + 0.5*y2
        c = y1

        # local → global mapping
        poly_vals = a*nodes**2 + b*nodes + c
        total += np.dot(weights, poly_vals)

    return total

class Line:
    """ A Line class used for convenient definition of integration lines"""
    def __init__(self, xpts: np.ndarray,
                 ypts: np.ndarray,
                 zpts: np.ndarray):
        self.xs: np.ndarray = xpts
        self.ys: np.ndarray = ypts
        self.zs: np.ndarray = zpts
        self.dxs: np.ndarray = xpts[1:] - xpts[:-1]
        self.dys: np.ndarray = ypts[1:] - ypts[:-1]
        self.dzs: np.ndarray = zpts[1:] - zpts[:-1]
        self.dl = np.sqrt(self.dxs**2 + self.dys**2 + self.dzs**2)
        self.length: float = np.sum(np.sqrt(self.dxs**2 + self.dys**2 + self.dzs**2))
        self.l: np.ndarray = np.concatenate((np.array([0,]), np.cumsum(self.dl))) 
        self.xmid: np.ndarray = 0.5*(xpts[:-1] + xpts[1:])
        self.ymid: np.ndarray = 0.5*(ypts[:-1] + ypts[1:])
        self.zmid: np.ndarray = 0.5*(zpts[:-1] + zpts[1:])

        self.dx = self.dxs[0]
        self.dy = self.dys[0]
        self.dz = self.dzs[0]
    
    @property
    def cmid(self) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        return self.xmid, self.ymid, self.zmid

    @property
    def cpoint(self) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        return self.xs, self.ys, self.zs
    
    @staticmethod
    def from_points(start: np.ndarray, end: np.ndarray, Npts: int) -> Line:
        x1, y1, z1 = start
        x2, y2, z2 = end
        xs = np.linspace(x1, x2, Npts)
        ys = np.linspace(y1, y2, Npts)
        zs = np.linspace(z1, z2, Npts)
        return Line(xs, ys, zs)
    
    def line_integral(self, evalfunc: Callable) -> complex:
        """Compute the line integral for a complex vector field function evalfunc."""
        Ex, Ey, Ez = evalfunc(*self.cpoint)
        EdotL = Ex*self.dx + Ey*self.dy + Ez*self.dz
        return gauss3_composite(self.l, EdotL)


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from numba_progress import ProgressBar
from ...mesh3d import SurfaceMesh
import numpy as np
from loguru import logger
from numba import c8, c16, f8, i8, njit, prange, typeof, f4
from numba.types import Tuple as TupleType
from numba_progress.progress import ProgressBarType

LR = 0.001
@njit(
    TupleType((c8[:, :], c8[:, :]))(
        c8[:, :],
        c8[:, :],
        f4[:, :],
        f4[:, :],
        f4[:, :],
        f4,
        ProgressBarType,
    ),
    parallel=True,
    fastmath=True,
    cache=True,
    nogil=True,
)
def stratton_chu_ff(Ein, Hin, vis, wns, tpout, k0, pgb):
    
    Ex = Ein[0, :].flatten()
    Ey = Ein[1, :].flatten()
    Ez = Ein[2, :].flatten()
    Hx = Hin[0, :].flatten()
    Hy = Hin[1, :].flatten()
    Hz = Hin[2, :].flatten()
    vx = vis[0, :].flatten()
    vy = vis[1, :].flatten()
    vz = vis[2, :].flatten()
    nx = wns[0, :].flatten()
    ny = wns[1, :].flatten()
    nz = wns[2, :].flatten()
    
    Emag = np.sqrt(np.abs(Ex)**2 + np.abs(Ey)**2 + np.abs(Ez)**2)
    
    Elevel = np.max(Emag) * LR
    ids = np.argwhere(Emag > Elevel)
    Nids = ids.shape[0]
    #iadd = NT // Nids
    Ex = Ex[Emag > Elevel]
    Ey = Ey[Emag > Elevel]
    Ez = Ez[Emag > Elevel]
    Hx = Hx[Emag > Elevel]
    Hy = Hy[Emag > Elevel]
    Hz = Hz[Emag > Elevel]
    vx = vx[Emag > Elevel]
    vy = vy[Emag > Elevel]
    vz = vz[Emag > Elevel]
    nx = nx[Emag > Elevel]
    ny = ny[Emag > Elevel]
    nz = nz[Emag > Elevel]

    thout = tpout[0, :]
    phout = tpout[1, :]

    rx = np.sin(thout) * np.cos(phout)
    ry = np.sin(thout) * np.sin(phout)
    rz = np.cos(thout)

    kx = k0 * rx
    ky = k0 * ry
    kz = k0 * rz

    N = tpout.shape[1]

    Eout = np.zeros((3, N)).astype(np.complex64)
    Hout = np.zeros((3, N)).astype(np.complex64)

    Eoutx = np.zeros((N,)).astype(np.complex64)
    Eouty = np.zeros((N,)).astype(np.complex64)
    Eoutz = np.zeros((N,)).astype(np.complex64)

    Z0 = np.float32(376.73031366857)

    Q = np.complex64(-1j * k0 / (4 * np.pi))
    ii = np.complex64(1j)

    NxHx = ny * Hz - nz * Hy
    NxHy = nz * Hx - nx * Hz
    NxHz = nx * Hy - ny * Hx

    NxEx = ny * Ez - nz * Ey
    NxEy = nz * Ex - nx * Ez
    NxEz = nx * Ey - ny * Ex

    for j in prange(Nids):
        xi = vx[j]
        yi = vy[j]
        zi = vz[j]
        G = np.exp(ii * (kx * xi + ky * yi + kz * zi))

        RxNxHx = ry * NxHz[j] - rz * NxHy[j]
        RxNxHy = rz * NxHx[j] - rx * NxHz[j]
        RxNxHz = rx * NxHy[j] - ry * NxHx[j]

        ie1x = (NxEx[j] - Z0 * RxNxHx) * G
        ie1y = (NxEy[j] - Z0 * RxNxHy) * G
        ie1z = (NxEz[j] - Z0 * RxNxHz) * G

        Eoutx += Q * (ry * ie1z - rz * ie1y)
        Eouty += Q * (rz * ie1x - rx * ie1z)
        Eoutz += Q * (rx * ie1y - ry * ie1x)

        # ii += iadd
        pgb.update(1)
    Eout[0, :] = Eoutx
    Eout[1, :] = Eouty
    Eout[2, :] = Eoutz

    Hout[0, :] = (ry * Eoutz - rz * Eouty) / Z0
    Hout[1, :] = (rz * Eoutx - rx * Eoutz) / Z0
    Hout[2, :] = (rx * Eouty - ry * Eoutx) / Z0

    return Eout, Hout

def stratton_chu(Ein, Hin, mesh: SurfaceMesh, theta: np.ndarray, phi: np.ndarray, k0: float):

    Ein = np.array(Ein)
    Hin = np.array(Hin)

    Emag = np.sqrt(np.abs(Ein[0,:])**2 + np.abs(Ein[1,:])**2 + np.abs(Ein[2,:])**2)
    Ntot = np.argwhere(Emag>0.000001*np.max(Emag)).shape[0]
    logger.debug(f'Percentage Included: {Ntot/Emag.shape[0]*100:.0f}%')
    areas = mesh.areas
    vis = mesh.edge_centers

    wns = np.zeros_like(vis).astype(np.float32)

    tri_normals = mesh.normals
    tri_ids = mesh.tri_to_edge

    for i in range(mesh.n_tris):
        n = tri_normals[:,i]
        i1, i2, i3 = tri_ids[:,i]
        wns[:,i1] += n*areas[i]/3
        wns[:,i2] += n*areas[i]/3
        wns[:,i3] += n*areas[i]/3
    
    Eout = None
    Hout = None
    tpout = np.array([theta, phi])
    with ProgressBar(total=Ntot, ncols=100, dynamic_ncols=False) as pgb:
        Eout, Hout = stratton_chu_ff(
            Ein.astype(np.complex64),
            Hin.astype(np.complex64),
            vis.astype(np.float32),
            wns.astype(np.float32),
            tpout.astype(np.float32),
            np.float32(k0),
            pgb,
        )
    return Eout.astype(np.complex128), Hout.astype(np.complex128)
